import unittest

from pyparsing import ParseException

from pynusmv_tools.ctlk.parsing import parseCTLK
from pynusmv_tools.ctlk.ast import (Spec, TrueExp, FalseExp, Init, Reachable,
                            Atom, Not, And, Or, Implies, Iff, 
                            AF, AG, AX, AU, AW, EF, EG, EX, EU, EW,
                            nK, nE, nD, nC, K, E, D, C)

class TestParsing(unittest.TestCase):
    
    def test_subclassing(self):
        self.assertTrue(issubclass(TrueExp, Spec))
        self.assertTrue(issubclass(FalseExp, Spec))
        self.assertTrue(issubclass(Init, Spec))
        self.assertTrue(issubclass(Reachable, Spec))
        self.assertTrue(issubclass(Atom, Spec))
        self.assertTrue(issubclass(Not, Spec))
        self.assertTrue(issubclass(And, Spec))
        self.assertTrue(issubclass(Or, Spec))
        self.assertTrue(issubclass(Implies, Spec))
        self.assertTrue(issubclass(Iff, Spec))
        self.assertTrue(issubclass(AF, Spec))
        self.assertTrue(issubclass(AG, Spec))
        self.assertTrue(issubclass(AX, Spec))
        self.assertTrue(issubclass(AU, Spec))
        self.assertTrue(issubclass(AW, Spec))
        self.assertTrue(issubclass(EF, Spec))
        self.assertTrue(issubclass(EG, Spec))
        self.assertTrue(issubclass(EX, Spec))
        self.assertTrue(issubclass(EU, Spec))
        self.assertTrue(issubclass(EW, Spec))
        self.assertTrue(issubclass(nK, Spec))
        self.assertTrue(issubclass(nE, Spec))
        self.assertTrue(issubclass(nD, Spec))
        self.assertTrue(issubclass(nC, Spec))
        self.assertTrue(issubclass(K, Spec))
        self.assertTrue(issubclass(E, Spec))
        self.assertTrue(issubclass(D, Spec))
        self.assertTrue(issubclass(C, Spec))
    
    def test_atom(self):
        s = "'c <= 3'"
        asts = parseCTLK(s)
        self.assertEqual(len(asts), 1)
        
        ast = asts[0]
        self.assertTrue(isinstance(ast, Spec))
        self.assertTrue(isinstance(ast, Atom))
        self.assertEqual(ast.value, "c <= 3")
        
    
    def test_fail(self):
        kos = ["", "'test", "A<'test'>H 'q'", "O", "A<'a'>E<'e'>G 'true'",
               "E<'ac'>[A<'ac'>['a' E<'ac'>['b' W 'c'] W 'd'] "
               "U A<'ac'>['e' U 'f']]", "A<'t't'>G 'q'"]
        
        for s in kos:
            with self.assertRaises(ParseException):
                parseCTLK(s)
                
                
    def test_true(self):
        s = "True"
        asts = parseCTLK(s)
        self.assertEqual(len(asts), 1)
        
        ast = asts[0]
        self.assertTrue(isinstance(ast, Spec))
        self.assertTrue(isinstance(ast, TrueExp))
        
        
    def test_false(self):
        s = "False"
        asts = parseCTLK(s)
        self.assertEqual(len(asts), 1)
        
        ast = asts[0]
        self.assertTrue(isinstance(ast, Spec))
        self.assertTrue(isinstance(ast, FalseExp))
        
    
    def test_init(self):
        s = "Init"
        asts = parseCTLK(s)
        self.assertEqual(len(asts), 1)
        
        ast = asts[0]
        self.assertTrue(isinstance(ast, Spec))
        self.assertTrue(isinstance(ast, Init))
        
    
    def test_reachable(self):
        s = "Reachable"
        asts = parseCTLK(s)
        self.assertEqual(len(asts), 1)
        
        ast = asts[0]
        self.assertTrue(isinstance(ast, Spec))
        self.assertTrue(isinstance(ast, Reachable))
    
    
    def test_not(self):
        s = "~'a'"
        asts = parseCTLK(s)
        self.assertEqual(len(asts), 1)
        
        ast = asts[0]
        self.assertTrue(isinstance(ast, Spec))
        self.assertTrue(isinstance(ast, Not))
        self.assertTrue(isinstance(ast.child, Spec))
        self.assertTrue(isinstance(ast.child, Atom))
        self.assertEqual(ast.child.value, "a")
        
    
    def test_and(self):
        s = "'a' & ('b' & 'c')"
        asts = parseCTLK(s)
        self.assertEqual(len(asts), 1)
        
        ast = asts[0]
        self.assertTrue(isinstance(ast, Spec))
        self.assertTrue(isinstance(ast, And))
        self.assertTrue(isinstance(ast.left, Spec))
        self.assertTrue(isinstance(ast.left, Atom))
        self.assertEqual(ast.left.value, "a")
        self.assertTrue(isinstance(ast.right, Spec))
        self.assertTrue(isinstance(ast.right, And))
        self.assertTrue(isinstance(ast.right.left, Spec))
        self.assertTrue(isinstance(ast.right.left, Atom))
        self.assertEqual(ast.right.left.value, "b")
        self.assertTrue(isinstance(ast.right.right, Spec))
        self.assertTrue(isinstance(ast.right.right, Atom))
        self.assertEqual(ast.right.right.value, "c")
        
    
    def test_implies(self):
        s = "'a' -> 'b'"
        asts = parseCTLK(s)
        self.assertEqual(len(asts), 1)
        
        ast = asts[0]
        self.assertTrue(isinstance(ast, Spec))
        self.assertTrue(isinstance(ast, Implies))
        self.assertTrue(isinstance(ast.left, Spec))
        self.assertTrue(isinstance(ast.left, Atom))
        self.assertEqual(ast.left.value, "a")
        self.assertTrue(isinstance(ast.right, Spec))
        self.assertTrue(isinstance(ast.right, Atom))
        self.assertEqual(ast.right.value, "b")
        
        
    def test_iff(self):
        s = "'a' <-> 'b = 3'"
        asts = parseCTLK(s)
        self.assertEqual(len(asts), 1)
        
        ast = asts[0]
        self.assertTrue(isinstance(ast, Spec))
        self.assertTrue(isinstance(ast, Iff))
        self.assertTrue(isinstance(ast.left, Spec))
        self.assertTrue(isinstance(ast.left, Atom))
        self.assertEqual(ast.left.value, "a")
        self.assertTrue(isinstance(ast.right, Spec))
        self.assertTrue(isinstance(ast.right, Atom))
        self.assertEqual(ast.right.value, "b = 3")
        
        
    def test_logicals(self):
        s = "'a' & 'c' -> (~'b' | 'c')"
        asts = parseCTLK(s)
        self.assertEqual(len(asts), 1)
        
        ast = asts[0]
        self.assertTrue(isinstance(ast, Spec))
        self.assertTrue(isinstance(ast, Implies))
        self.assertTrue(isinstance(ast.left, Spec))
        self.assertTrue(isinstance(ast.left, And))
        self.assertTrue(isinstance(ast.left.left, Spec))
        self.assertTrue(isinstance(ast.left.left, Atom))
        self.assertEqual(ast.left.left.value, "a")
        self.assertTrue(isinstance(ast.left.right, Spec))
        self.assertTrue(isinstance(ast.left.right, Atom))
        self.assertEqual(ast.left.right.value, "c")
        self.assertTrue(isinstance(ast.right, Spec))
        self.assertTrue(isinstance(ast.right, Or))
        self.assertTrue(isinstance(ast.right.left, Spec))
        self.assertTrue(isinstance(ast.right.left, Not))
        self.assertTrue(isinstance(ast.right.left.child, Spec))
        self.assertTrue(isinstance(ast.right.left.child, Atom))
        self.assertEqual(ast.right.left.child.value, "b")
        self.assertTrue(isinstance(ast.right.right, Spec))
        self.assertTrue(isinstance(ast.right.right, Atom))
        self.assertEqual(ast.right.right.value, "c")
        
        
    def test_ex(self):
        s = "EX 'a'"
        asts = parseCTLK(s)
        self.assertEqual(len(asts), 1)
        
        ast = asts[0]
        self.assertEqual(type(ast), EX)
        self.assertEqual(type(ast.child), Atom)
        self.assertEqual(ast.child.value, "a")
        
        
    def test_ex_and(self):
        s = "EX 'a' & 'b'"
        asts = parseCTLK(s)
        self.assertEqual(len(asts), 1)
        
        ast = asts[0]
        self.assertEqual(type(ast), And)
        self.assertEqual(type(ast.left), EX)
        self.assertEqual(type(ast.left.child), Atom)
        self.assertEqual(ast.left.child.value, "a")
        self.assertEqual(type(ast.right), Atom)
        self.assertEqual(ast.right.value, "b")
        
    
    def test_eu(self):
        s = "E['r' & 'd' U 's' | EF 'q']"
        asts = parseCTLK(s)
        self.assertEqual(len(asts), 1)
        
        ast = asts[0]
        self.assertEqual(type(ast), EU)
        self.assertEqual(type(ast.left), And)
        self.assertEqual(type(ast.left.left), Atom)
        self.assertEqual(ast.left.left.value, "r")
        self.assertEqual(type(ast.left.right), Atom)
        self.assertEqual(ast.left.right.value, "d")
        self.assertEqual(type(ast.right), Or)
        self.assertEqual(type(ast.right.left), Atom)
        self.assertEqual(ast.right.left.value, "s")
        self.assertEqual(type(ast.right.right), EF)
        self.assertEqual(type(ast.right.right.child), Atom)
        self.assertEqual(ast.right.right.child.value, "q")
        
        
    def test_eu_d(self):
        s = "E[D<'a'>'r' & 'd' U 's' | nD<'b','c'> 'q']"
        asts = parseCTLK(s)
        self.assertEqual(len(asts), 1)
        
        ast = asts[0]
        self.assertEqual(type(ast), EU)
        self.assertEqual(type(ast.left), And)
        self.assertEqual(type(ast.left.left), D)
        self.assertEqual(type(ast.left.left.group), list)
        self.assertEqual(len(ast.left.left.group), 1)
        self.assertEqual(type(ast.left.left.group[0]), Atom)
        self.assertEqual(ast.left.left.group[0].value, "a")
        self.assertEqual(type(ast.left.right), Atom)
        self.assertEqual(ast.left.right.value, "d")        
        self.assertEqual(type(ast.right), Or)
        self.assertEqual(type(ast.right.left), Atom)
        self.assertEqual(ast.right.left.value, "s")
        self.assertEqual(type(ast.right.right), nD)
        self.assertEqual(type(ast.right.right.group), list)
        self.assertEqual(len(ast.right.right.group), 2)
        self.assertEqual(type(ast.right.right.group[0]), Atom)
        self.assertEqual(ast.right.right.group[0].value, "b")
        self.assertEqual(type(ast.right.right.group[1]), Atom)
        self.assertEqual(ast.right.right.group[1].value, "c")
        self.assertEqual(type(ast.right.right.child), Atom)
        self.assertEqual(ast.right.right.child.value, "q")
        
        
    def test_k(self):
        s = "K<'c'> 'a'"
        asts = parseCTLK(s)
        self.assertEqual(len(asts), 1)
        
        ast = asts[0]
        self.assertEqual(type(ast), K)
        self.assertEqual(type(ast.agent), Atom)
        self.assertEqual(ast.agent.value, "c")
        self.assertEqual(type(ast.child), Atom)
        self.assertEqual(ast.child.value, "a")
        
    
    def test_nk_and(self):
        s = "nK<'a'> 'q' & 'r'"
        asts = parseCTLK(s)
        self.assertEqual(len(asts), 1)
        
        ast = asts[0]
        self.assertEqual(type(ast), And)
        self.assertEqual(type(ast.left), nK)
        self.assertEqual(type(ast.left.agent), Atom)
        self.assertEqual(ast.left.agent.value, "a")
        self.assertEqual(type(ast.left.child), Atom)
        self.assertEqual(ast.left.child.value, "q")
        self.assertEqual(type(ast.right), Atom)
        self.assertEqual(ast.right.value, "r")
        
        
    def test_weird(self):
        s = """
            AF (
                (K<'c1'> (AF (K<'c1'> 'c2.payer' | K<'c1'> ~'c2.payer')))
                |
                (K<'c1'> (~AF (K<'c1'> 'c2.payer' | K<'c1'> ~'c2.payer')))
               )
            """
        asts = parseCTLK(s)
        self.assertEqual(len(asts), 1)
        
        ast = asts[0]
        self.assertEqual(type(ast), AF)
        self.assertEqual(type(ast.child), Or)
        self.assertEqual(type(ast.child.left), K)
        self.assertEqual(type(ast.child.left.agent), Atom)
        self.assertEqual(ast.child.left.agent.value, "c1")
        self.assertEqual(type(ast.child.left.child), AF)
        self.assertEqual(type(ast.child.left.child.child), Or)
        self.assertEqual(type(ast.child.left.child.child.left), K)
        self.assertEqual(type(ast.child.left.child.child.right), K)
        self.assertEqual(type(ast.child.left.child.child.right.agent), Atom)
        self.assertEqual(ast.child.left.child.child.right.agent.value, "c1")
        self.assertEqual(type(ast.child.left.child.child.right.child), Not)
        self.assertEqual(
                       type(ast.child.left.child.child.right.child.child), Atom)
        self.assertEqual(
                 ast.child.left.child.child.right.child.child.value, "c2.payer")
        self.assertEqual(type(ast.child.right), K)
        
    
    def test_epistemic_temporal_logical(self):
        s = "nK<'ag'> AX ~'b'"
        asts = parseCTLK(s)
        self.assertEqual(len(asts), 1)
        
        ast = asts[0]
        self.assertEqual(type(ast), nK)
        self.assertEqual(type(ast.agent), Atom)
        self.assertEqual(ast.agent.value, "ag")
        self.assertEqual(type(ast.child), AX)
        self.assertEqual(type(ast.child.child), Not)
        self.assertEqual(type(ast.child.child.child), Atom)
        self.assertEqual(ast.child.child.child.value, "b")
        
    
    def test_full(self):
        s = """
            A[
                K<'a'> EF Init
                &
                AG E<'a','b'> (True | 'r')
            U
                nC<'c','a'> E['s' W False -> AX 's']
            ] -> Reachable
            """
        asts = parseCTLK(s)
        self.assertEqual(len(asts), 1)
        
        ast = asts[0]
        self.assertEqual(type(ast), Implies)
        self.assertEqual(type(ast.left), AU)
        self.assertEqual(type(ast.left.left), And)
        
        # K<'a'> EF Init
        self.assertEqual(type(ast.left.left.left), K)
        self.assertEqual(type(ast.left.left.left.agent), Atom)
        self.assertEqual(ast.left.left.left.agent.value, "a")
        self.assertEqual(type(ast.left.left.left.child), EF)
        self.assertEqual(type(ast.left.left.left.child.child), Init)
        
        # AG E<'a','b'> (True | 'r')
        self.assertEqual(type(ast.left.left.right), AG)
        self.assertEqual(type(ast.left.left.right.child), E)
        self.assertEqual(type(ast.left.left.right.child.group), list)
        self.assertEqual(len(ast.left.left.right.child.group), 2)
        self.assertEqual(type(ast.left.left.right.child.group[0]), Atom)
        self.assertEqual(ast.left.left.right.child.group[0].value, "a")
        self.assertEqual(type(ast.left.left.right.child.group[1]), Atom)
        self.assertEqual(ast.left.left.right.child.group[1].value, "b")
        self.assertEqual(type(ast.left.left.right.child.child), Or)
        self.assertEqual(type(ast.left.left.right.child.child.left), TrueExp)
        self.assertEqual(type(ast.left.left.right.child.child.right), Atom)
        self.assertEqual(ast.left.left.right.child.child.right.value, "r")
                
        # nC<'c','a'> E['s' W False -> AX 's']
        self.assertEqual(type(ast.left.right), nC)
        self.assertEqual(type(ast.left.right.group), list)
        self.assertEqual(len(ast.left.right.group), 2)
        self.assertEqual(type(ast.left.right.group[0]), Atom)
        self.assertEqual(ast.left.right.group[0].value, "c")
        self.assertEqual(type(ast.left.right.group[1]), Atom)
        self.assertEqual(ast.left.right.group[1].value, "a")
        self.assertEqual(type(ast.left.right.child), EW)
        self.assertEqual(type(ast.left.right.child.left), Atom)
        self.assertEqual(ast.left.right.child.left.value, "s")
        self.assertEqual(type(ast.left.right.child.right), Implies)
        self.assertEqual(type(ast.left.right.child.right.left), FalseExp)
        self.assertEqual(type(ast.left.right.child.right.right), AX)
        self.assertEqual(type(ast.left.right.child.right.right.child), Atom)
        self.assertEqual(ast.left.right.child.right.right.child.value, "s")        
        
        self.assertEqual(type(ast.right), Reachable)


"""
This module contains a configuration variable accessible from everywhere
in the ATLK_irF model-checking tool used to tune the behaviour of different
algorithms.
config is the main object containing configuration variables, and these
variables are accessible through attributes of config.


The list of parameters follows. Every indentation is a sub-parameter;
for example, config.partial.early.type is type under early under partial.
Top-level parameters are sub-parameters of config.

 debug (boolean): printing or not debugging information
 partial (set of parameters): set of parameters for partial strategies
    early: set of parameters for early termination of partial strategies
        type (string): type of early termination
            - None: no early termination
            - full: stop when all Z states are in sat
            - partial: whenever sat grows, recompute strategies for the rest
            - threshold: when remaining states decrease of threshold,
                         recompute the strategies
        threshold (int): the threshold for early termination with threshold
    caching (boolean): whether or not caching states satisfying or not
                       sub-formulas
    filtering (boolean): activate filtering
    separation: separating Z states into sub-sets to check one at a time
        type (string): type of separation
            - None: no separation
            - random: pick a random state in Z and take its equivalence class
            - reach: pick first reachable state from initial ones and take
                     its equivalence class
    garbage: explicit call to garbage collection
        type (string): type to garbage collection calls
            - None: no explicit call to garbage collection
            - each: collect after every strategy
            - step: collect every step strategies
        step (int): number of strategies between each collection when type is
                    step
"""

class AttrDict(dict):
    def __init__(self, *args, **kwargs):
        super(AttrDict, self).__init__(*args, **kwargs)
        self.__dict__ = self

config = AttrDict()



# -------------
# General stuff
# -------------

# Whether or not printing debug information at console
config.debug = False

# Garbage collection
# ------------------

config.garbage = AttrDict()

# Which technique to use
#   None: garbage collection is never called explicitely
#   each: garbage collection is called after each strategy
#   step: garbage collection is called after the corresponding number of 
#         strategies
config.garbage.type = None

# Number of strategies to check before calling again garbage collection
config.garbage.step = 100



# ---------------------------------------------------
# Partial strategies implementation related variables
# ---------------------------------------------------
config.partial = AttrDict()


# Early termination in partial strategies
# ---------------------------------------
config.partial.early = AttrDict()
# Early termination type used
#   None: early termination not used
#   full: when all the states satisfy the property
#   partial: whenever sat grows, recompute the strategies with remaining states
#   threshold: when remaining states decreases below the given threshold, 
#              recompute the strategies for these remaining states
config.partial.early.type = None
# Early termination recomputation threshold: what is the factor between
# current remaining below which we have to recompute the strategies for 
# remaining states
config.partial.early.threshold = 0.75


# Caching subformulas in partial strategies
# -----------------------------------------

# Whether or not perform caching through accumulation of truth values
config.partial.caching = False


# Pre-filtering out losing moves
# ------------------------------

# Whether or not pre-filter losing moves
# before splitting into partial strategies
config.partial.filtering = False


# States set separation
# ---------------------

config.partial.separation = AttrDict()

# Which technique to use
#   None: states set separation not used
#   random: a random state is picked, its equivalence class is chosen
#   reach: the first reached state, from initial states,
#          and its equivalence class are chosen
config.partial.separation.type = None


# Alternating variant:
# --------------------

config.partial.alternate = AttrDict()

# Which technique to use for alternation
#   univ: perform universal filtering
#   strat: perform full-observability-based filtering
# This parameter should contain at least one of the two values
config.partial.alternate.type = {"univ", "strat"}


"""
AST classes for ATLK formulas representation.
"""

from collections import namedtuple

class Spec:
    """A Spec is represents a ATLK formula."""
    def subformulas(self):
        """Return the set of sub-formulas of this formula, including itself."""
        raise NotImplementedError("Should be implemented by subclasses.")

# TrueExp, FalseExp, Init, Reachable
class TrueExp(Spec):
    def __str__(self):
        return "True"
    def subformulas(self):
        return {self}
        
class FalseExp(Spec):
    def __str__(self):
        return "False"
    def subformulas(self):
        return {self}
        
class Init(Spec):
    def __str__(self):
        return "Init"
    def subformulas(self):
        return {self}
        
class Reachable(Spec):
    def __str__(self):
        return "Reachable"
    def subformulas(self):
        return {self}
        
class Atom(Spec):
    def __init__(self, value):
        self.value = value
    def __str__(self):
        return "'" + str(self.value) + "'"
    def subformulas(self):
        return {self}

# logical : Not, And, Or, Implies, Iff
class Not(Spec):
    def __init__(self, child):
        self.child = child
    def __str__(self):
        return "~" + "(" + str(self.child) + ")"
    def subformulas(self):
        return {self} | self.child.subformulas()

class And(Spec):
    def __init__(self, left, right):
        self.left = left
        self.right = right
    def __str__(self):
        return "(" + str(self.left) + " & " + str(self.right) + ")"
    def subformulas(self):
        return {self} | self.left.subformulas() | self.right.subformulas()

class Or(Spec):
    def __init__(self, left, right):
        self.left = left
        self.right = right
    def __str__(self):
        return "(" + str(self.left) + " | " + str(self.right) + ")"
    def subformulas(self):
        return {self} | self.left.subformulas() | self.right.subformulas()

class Implies(Spec):
    def __init__(self, left, right):
        self.left = left
        self.right = right
    def __str__(self):
        return "(" + str(self.left) + " -> " + str(self.right) + ")"
    def subformulas(self):
        return {self} | self.left.subformulas() | self.right.subformulas()

class Iff(Spec):
    def __init__(self, left, right):
        self.left = left
        self.right = right
    def __str__(self):
        return "(" + str(self.left) + " <-> " + str(self.right) + ")"
    def subformulas(self):
        return {self} | self.left.subformulas() | self.right.subformulas()
                                  

# temporal : AF, AG, AX, AU, AW, EF, EG, EX, EU, EW
class AF(Spec):
    def __init__(self, child):
        self.child = child
    def __str__(self):
        return  "A" + "F " + str(self.child)
    def subformulas(self):
        return {self} | self.child.subformulas()

class AG(Spec):
    def __init__(self, child):
        self.child = child
    def __str__(self):
        return  "A" + "G " + str(self.child)
    def subformulas(self):
        return {self} | self.child.subformulas()

class AX(Spec):
    def __init__(self, child):
        self.child = child
    def __str__(self):
        return  "A" + "X " + str(self.child)
    def subformulas(self):
        return {self} | self.child.subformulas()

class AU(Spec):
    def __init__(self, left, right):
        self.left = left
        self.right = right
    def __str__(self):
        return ("A" + "[" + str(self.left) +
                      " U " + str(self.right) + "]")
    def subformulas(self):
        return {self} | self.left.subformulas() | self.right.subformulas()

class AW(Spec):
    def __init__(self, left, right):
        self.left = left
        self.right = right
    def __str__(self):
        return ("A" + "[" + str(self.left) +
                      " W " + str(self.right) + "]")
    def subformulas(self):
        return {self} | self.left.subformulas() | self.right.subformulas()

class EF(Spec):
    def __init__(self, child):
        self.child = child
    def __str__(self):
        return  "E" + "F " + str(self.child)
    def subformulas(self):
        return {self} | self.child.subformulas()

class EG(Spec):
    def __init__(self, child):
        self.child = child
    def __str__(self):
        return  "E" + "G " + str(self.child)
    def subformulas(self):
        return {self} | self.child.subformulas()

class EX(Spec):
    def __init__(self, child):
        self.child = child
    def __str__(self):
        return  "E" + "X " + str(self.child)
    def subformulas(self):
        return {self} | self.child.subformulas()

class EU(Spec):
    def __init__(self, left, right):
        self.left = left
        self.right = right
    def __str__(self):
        return ("E" + "[" + str(self.left) +
                      " U " + str(self.right) + "]")
    def subformulas(self):
        return {self} | self.left.subformulas() | self.right.subformulas()

class EW(Spec):
    def __init__(self, left, right):
        self.left = left
        self.right = right
    def __str__(self):
        return ("E" + "[" + str(self.left) +
                      " W " + str(self.right) + "]")
    def subformulas(self):
        return {self} | self.left.subformulas() | self.right.subformulas()
                                

# epistemic : nK, nE, nD, nC, K, E, D, C
class nK(Spec):
    def __init__(self, agent, child):
        self.agent = agent
        self.child = child
    def __str__(self):
        return  ("nK" + "<" + str(self.agent) + ">" + " " + str(self.child))
    def subformulas(self):
        return {self} | self.child.subformulas()
        
class nE(Spec):
    def __init__(self, group, child):
        self.group = group
        self.child = child
    def __str__(self):
        return  ("nE" + "<" + ','.join([ag.value for ag in self.group]) +
                              ">" + " " + str(self.child))
    def subformulas(self):
        return {self} | self.child.subformulas()
        
class nD(Spec):
    def __init__(self, group, child):
        self.group = group
        self.child = child
    def __str__(self):
        return  ("nD" + "<" + ','.join([ag.value for ag in self.group]) +
                              ">" + " " + str(self.child))
    def subformulas(self):
        return {self} | self.child.subformulas()
        
class nC(Spec):
    def __init__(self, group, child):
        self.group = group
        self.child = child
    def __str__(self):
        return  ("nC" + "<" + ','.join([ag.value for ag in self.group]) +
                              ">" + " " + str(self.child))
    def subformulas(self):
        return {self} | self.child.subformulas()

class K(Spec):
    def __init__(self, agent, child):
        self.agent = agent
        self.child = child
    def __str__(self):
        return  ("K" + "<" + str(self.agent) + ">" + " " + str(self.child))
    def subformulas(self):
        return {self} | self.child.subformulas()
        
class E(Spec):
    def __init__(self, group, child):
        self.group = group
        self.child = child
    def __str__(self):
        return  ("E" + "<" + ','.join([ag.value for ag in self.group]) +
                              ">" + " " + str(self.child))
    def subformulas(self):
        return {self} | self.child.subformulas()
        
class D(Spec):
    def __init__(self, group, child):
        self.group = group
        self.child = child
    def __str__(self):
        return  ("D" + "<" + ','.join([ag.value for ag in self.group]) +
                              ">" + " " + str(self.child))
    def subformulas(self):
        return {self} | self.child.subformulas()
        
class C(Spec):
    def __init__(self, group, child):
        self.group = group
        self.child = child
    def __str__(self):
        return  ("C" + "<" + ','.join([ag.value for ag in self.group]) +
                              ">" + " " + str(self.child))
    def subformulas(self):
        return {self} | self.child.subformulas()


# strategic : <g>F, <g>G, <g>X, <g>U, <g>W, [g]F, [g]G, [g]X, [g]U, [g]W
class CEF(Spec):
    def __init__(self, group, child):
        self.group = group
        self.child = child
    def __str__(self):
        return ("<" +','.join([ag.value for ag in self.group])+ ">"+
                "F " + str(self.child))
    def subformulas(self):
        return {self} | self.child.subformulas()

class CEG(Spec):
    def __init__(self, group, child):
        self.group = group
        self.child = child
    def __str__(self):
        return ("<" +','.join([ag.value for ag in self.group])+ ">"+
                "G " + str(self.child))
    def subformulas(self):
        return {self} | self.child.subformulas()

class CEX(Spec):
    def __init__(self, group, child):
        self.group = group
        self.child = child
    def __str__(self):
        return ("<" +','.join([ag.value for ag in self.group])+ ">"+
                "X " + str(self.child))
    def subformulas(self):
        return {self} | self.child.subformulas()

class CEU(Spec):
    def __init__(self, group, left, right):
        self.group = group
        self.left = left
        self.right = right
    def __str__(self):
        return ("<" +','.join([ag.value for ag in self.group])+ ">"+ 
                            "[" + str(self.left) +
                            " U " + str(self.right) + "]")
    def subformulas(self):
        return {self} | self.left.subformulas() | self.right.subformulas()

class CEW(Spec):
    def __init__(self, group, left, right):
        self.group = group
        self.left = left
        self.right = right
    def __str__(self):
        return ("<" +','.join([ag.value for ag in self.group])+ ">"+ 
                            "[" + str(self.left) +
                            " W " + str(self.right) + "]")
    def subformulas(self):
        return {self} | self.left.subformulas() | self.right.subformulas()

class CAF(Spec):
    def __init__(self, group, child):
        self.group = group
        self.child = child
    def __str__(self):
        return ("[" +','.join([ag.value for ag in self.group])+ "]"+
                "F " + str(self.child))
    def subformulas(self):
        return {self} | self.child.subformulas()

class CAG(Spec):
    def __init__(self, group, child):
        self.group = group
        self.child = child
    def __str__(self):
        return ("[" +','.join([ag.value for ag in self.group])+ "]"+
                "G " + str(self.child))
    def subformulas(self):
        return {self} | self.child.subformulas()

class CAX(Spec):
    def __init__(self, group, child):
        self.group = group
        self.child = child
    def __str__(self):
        return ("[" +','.join([ag.value for ag in self.group])+ "]"+
                "X " + str(self.child))
    def subformulas(self):
        return {self} | self.child.subformulas()

class CAU(Spec):
    def __init__(self, group, left, right):
        self.group = group
        self.left = left
        self.right = right
    def __str__(self):
        return ("[" +','.join([ag.value for ag in self.group])+ "]"+ 
                            "[" + str(self.left) +
                            " U " + str(self.right) + "]")
    def subformulas(self):
        return {self} | self.left.subformulas() | self.right.subformulas()

class CAW(Spec):
    def __init__(self, group, left, right):
        self.group = group
        self.left = left
        self.right = right
    def __str__(self):
        return ("[" +','.join([ag.value for ag in self.group])+ "]"+ 
                            "[" + str(self.left) +
                            " W " + str(self.right) + "]")
    def subformulas(self):
        return {self} | self.left.subformulas() | self.right.subformulas()


class Explanation():
    """
    Explanations consistute a graph explaining why a particular state satisfies
    an ATL property.
    
    An Explanation is composed of state and a list of successors reachable
    through inputs.
    """
    def __init__(self, state, successors=None):
        """
        state -- the state of the explanation node;
        successors -- a set of (inputs, successor) pairs where successors
                      are also explanations.
        """
        self.state = state
        self.successors = successors if successors else set()
        
    
    def dot(self):
        """
        Return a dot representation (as a text) of this explanation.
        """
        # Get all states, keep them and mark them
        ids = {}
        curid = 0
        extract = {self}
        while len(extract) > 0:
            expl = extract.pop()
            ids[expl] = "s" + str(curid)
            curid += 1
            for (action, succ) in expl.successors:
                if succ not in ids:
                    extract.add(succ)
                
        dot = "digraph {"
        
        # Add states to the dot representation
        for expl in ids:
            dot += (ids[expl] + " " + "[label=\"" +
                    '\\n'.join(var + "=" + val for var, val in
                                            expl.state.get_str_values().items()) +
                    "\"]" + ";\n")
        
        # For each state, add each transition to the representation
        for expl in ids:
            for action, succ in expl.successors:
                dot += (ids[expl] + "->" + ids[succ] + " " +
                        "[label=\"" + "\\n".join(var + "=" + val for var, val in 
                                        action.get_str_values().items()) + "\"]"
                        + ";\n")
        
        dot += "}"
        
        return dot


from pynusmv.mc import eval_ctl_spec
from .explain import explain

def check(fsm, spec):
    """
    Check whether fsm satisfies spec or not.
    
    Return (True, None) if spec is satisfied,
    and (False, cntex) otherwise,
    where cntex is TLACE node explaining the violation.
    """
    
    initbdd = fsm.init
    specbdd = eval_ctl_spec(fsm, spec)
    
    # Get violating states
    violating = initbdd & ~specbdd
    
    # If some initial states are not in specbdd, the spec if violated
    if violating.isnot_false():
        # Compute a counter-example
        state = fsm.pick_one_state(violating)
        return (False, explain(fsm, state, spec))
        
    # Otherwise, it is satisfied
    else:
        return (True, None)



import unittest

from pynusmv.fsm import BddFsm
from pynusmv.init import init_nusmv, deinit_nusmv

from pynusmv_tools.arctl.parsing import parseArctl
from pynusmv_tools.arctl.eval import evalArctl, evalArctl_from_string as evalStr
from pynusmv_tools.arctl.explain import (explain_eax, explain_eau, explain_eag,
                                 explain_witness)


class TestExplain(unittest.TestCase):
    
    def setUp(self):
        init_nusmv()
    
    def tearDown(self):
        deinit_nusmv()
    
    
    def init_model(self):
        fsm = BddFsm.from_filename("tests/pynusmv_tools/arctl/counters.smv")
        self.assertIsNotNone(fsm)
        return fsm
        
    def test_eax_explain(self):
        fsm = self.init_model()
        self.assertIsNotNone(fsm)
        
        spec = parseArctl("E<'TRUE'>X 'c1.c = 1'")[0]
        specbdd = evalArctl(fsm, spec)
        
        s = fsm.pick_one_state(specbdd & fsm.init)
        self.assertIsNotNone(s)
        
        ac = evalStr(fsm, "'TRUE'")
        phi = evalStr(fsm, "'c1.c = 1'")
        
        path = explain_eax(fsm, s, ac, phi)
        
        self.assertTrue(s == path[0])
        self.assertTrue(path[1] <= ac)
        self.assertTrue(path[2] <= phi)
        self.assertTrue(path[2] <= fsm.post(s, path[1]))
       
       
    def test_eau_explain(self):
        fsm = self.init_model()
        self.assertIsNotNone(fsm)
        
        spec = parseArctl("E<'TRUE'>['c1.c < 2' U 'c1.c = 2']")[0]
        specbdd = evalArctl(fsm, spec)
        
        s = fsm.pick_one_state(specbdd & fsm.init)
        self.assertTrue((specbdd & fsm.init).isnot_false())
        self.assertIsNotNone(s)
        self.assertTrue(s.isnot_false())
        
        ac = evalStr(fsm, "'TRUE'")
        phi = evalStr(fsm, "'c1.c < 2'")
        psi = evalStr(fsm, "'c1.c = 2'")
        
        path = explain_eau(fsm, s, ac, phi, psi)
        
        # Check that path contains states of phi and psi
        self.assertTrue(s == path[0])
        for state, inputs in zip(path[::2], path[1::2]):
            self.assertTrue(state <= phi)
            self.assertTrue(inputs <= ac)
        self.assertTrue(path[-1] <= psi)
        
        # Check that path is a path of fsm
        for s, i, sp in zip(path[::2], path[1::2], path[2::2]):
            self.assertTrue(sp <= fsm.post(s, i))
            
            
    def test_eag_explain(self):
        fsm = self.init_model()
        self.assertIsNotNone(fsm)
        
        spec = parseArctl("E<'run = rc1'>G 'c1.c <= 2'")[0]
        specbdd = evalArctl(fsm, spec)
        
        s = fsm.pick_one_state(specbdd & fsm.init)
        self.assertTrue((specbdd & fsm.init).isnot_false())
        self.assertIsNotNone(s)
        self.assertTrue(s.isnot_false())
        
        ac = evalStr(fsm, "'run = rc1'")
        phi = evalStr(fsm, "'c1.c <= 2'")
        
        (path, (inputs, loop)) = explain_eag(fsm, s, ac, phi)
        
        # loop and inputs are not None: the explanation is an infinite path
        self.assertIsNotNone(inputs)
        self.assertIsNotNone(loop)
        
        # Check that path contains states of phi and psi
        self.assertTrue(s == path[0])
        for state, inputs in zip(path[::2], path[1::2]):
            self.assertTrue(state <= phi)
            self.assertTrue(inputs <= ac)
        self.assertTrue(path[-1] <= phi)
        
        # Check that path is a path of fsm
        for s, i, sp in zip(path[::2], path[1::2], path[2::2]):
            self.assertTrue(sp <= fsm.post(s, i))
            
        # Check that loop is effectively a loop
        self.assertTrue(loop in path)
        self.assertTrue(loop <= fsm.post(path[-1], inputs))
        
        
    def test_explain_atom(self):
        fsm = self.init_model()
        self.assertIsNotNone(fsm)
        
        spec = parseArctl("'c1.c <= 2'")[0]
        specbdd = evalArctl(fsm, spec)
        
        s = fsm.pick_one_state(specbdd & fsm.init)
        self.assertTrue(s.isnot_false())
        
        (path, (inp, loop)) = explain_witness(fsm, s, spec)
        # Check that no loop
        self.assertIsNone(inp)
        self.assertIsNone(loop)
        
        # Check path
        self.assertEqual(path, (s,))
        
        
    def test_explain_and_atom(self):
        fsm = self.init_model()
        self.assertIsNotNone(fsm)
        
        spec = parseArctl("'c1.c <= 2' & 'c2.c < 1'")[0]
        specbdd = evalArctl(fsm, spec)
        
        s = fsm.pick_one_state(specbdd & fsm.init)
        self.assertTrue(s.isnot_false())
        
        (path, (inp, loop)) = explain_witness(fsm, s, spec)
        # Check that no loop
        self.assertIsNone(inp)
        self.assertIsNone(loop)
        
        # Check path
        self.assertEqual(path, (s,))
        


import unittest

from pynusmv.dd import BDD
from pynusmv.init import init_nusmv, deinit_nusmv
from pynusmv.mc import eval_simple_expression

from pynusmv.utils import fixpoint as fp

from pynusmv_tools.mas import glob

from pynusmv_tools.atl.check import check
from pynusmv_tools.atl.eval import evalATL, cex
from pynusmv_tools.atl.parsing import parseATL
from pynusmv_tools.atl.explain import (explain_cex, explain_cax, explain_ceu,
                               explain_cau, explain_ceg, explain_cag,
                               explain_cew, explain_caw)


class TestCheck(unittest.TestCase):
    
    def setUp(self):
        init_nusmv()
    
    def tearDown(self):
        glob.reset_globals()
        deinit_nusmv()
    
    
    def small(self):
        glob.load_from_file("tests/pynusmv_tools/atl/models/small-game.smv")
        fsm = glob.mas()
        self.assertIsNotNone(fsm)
        return fsm
        
    def cardgame(self):
        glob.load_from_file("tests/pynusmv_tools/atl/models/cardgame.smv")
        fsm = glob.mas()
        self.assertIsNotNone(fsm)
        return fsm
        
    def transmission(self):
        glob.load_from_file("tests/pynusmv_tools/atl/models/transmission.smv")
        fsm = glob.mas()
        self.assertIsNotNone(fsm)
        return fsm
    
    def tictactoe(self):
        glob.load_from_file("tests/pynusmv_tools/atl/models/tictactoe.smv")
        fsm = glob.mas()
        self.assertIsNotNone(fsm)
        return fsm
        
    def show_si(self, fsm, bdd):
        if bdd.isnot_false():
            sis = fsm.pick_all_states_inputs(bdd)
            print("SI count:", len(sis))
            for si in sis:
                print(si.get_str_values())
            
    def show_s(self, fsm, bdd):
        if bdd.isnot_false():
            ss = fsm.pick_all_states(bdd)
            print("S count:", len(ss))
            for s in ss:
                print(s.get_str_values())
            
    def show_i(self, fsm, bdd):
        if bdd.isnot_false():
            ii = fsm.pick_all_inputs(bdd)
            print("I count:", len(ii))
            for i in ii:
                print(i.get_str_values())
                
                
    def show_cex(self, explanation, spec):
        """Show the explanation of spec satisfied by explanation.state."""
        print()
        print("Explaining", spec)
        print(explanation.state.get_str_values())
        for action, succ in explanation.successors:
            print("action:", action.get_str_values())
            print("successor:", succ.state.get_str_values())
    
    
    def check_cex(self, fsm, explanation, agents, phi):
        """Check that the explanation is correct."""
        # Get the cubes
        gamma_cube = fsm.inputs_cube_for_agents(agents)
        ngamma_cube = fsm.bddEnc.inputsCube - gamma_cube
        
        # The first state satisfies the spec
        self.assertTrue(explanation.state <= cex(fsm, agents, phi))
        acts = BDD.false(fsm.bddEnc.DDmanager)
        states = BDD.false(fsm.bddEnc.DDmanager)
        for (act, succ) in explanation.successors:
            ag_action = act.forsome(ngamma_cube)
            # The successor satisfies phi
            self.assertTrue(succ.state <= phi)
            # The action is effectively possible
            self.assertTrue(act <= fsm.get_inputs_between_states(
                                                 explanation.state, succ.state))
            # Accumulate states and actions
            acts = acts | act
            states = states | succ.state
        
        # The reached states are effectively the reachable states
        # through the action
        self.assertTrue(states <= fsm.post(explanation.state, ag_action))
        self.assertTrue(states >= fsm.post(explanation.state, ag_action)
                                                        & fsm.bddEnc.statesMask)
                                                        
        # The actions are effectively all the possible actions completing ag_act
        self.assertTrue(acts <= ag_action)
        self.assertTrue(acts >= ag_action & 
                      fsm.get_inputs_between_states(explanation.state, states) &
                      fsm.bddEnc.inputsMask)
                      
                      
    def check_cax(self, fsm, explanation, agents, phi):
        """Check that the explanation is correct."""
        # Get the cubes
        gamma_cube = fsm.inputs_cube_for_agents(agents)
        ngamma_cube = fsm.bddEnc.inputsCube - gamma_cube
        
        # The first state satisfies the spec
        self.assertTrue(explanation.state <= ~cex(fsm, agents, ~phi))
        acts = BDD.false(fsm.bddEnc.DDmanager)
        states = BDD.false(fsm.bddEnc.DDmanager)
        for (act, succ) in explanation.successors:
            # The successor satisfies phi
            self.assertTrue(succ.state <= phi)
            # The action is effectively possible
            self.assertTrue(act <= fsm.get_inputs_between_states(
                                                 explanation.state, succ.state))
            # Accumulate states and actions
            acts = acts | act
            states = states | succ.state
        
        # The actions are effectively all the possible ones
        self.assertEqual((fsm.protocol(agents) &
                          explanation.state).forsome(fsm.bddEnc.statesCube).
                          forsome(ngamma_cube) &
                         fsm.bddEnc.statesMask,
                         acts.forsome(ngamma_cube) & fsm.bddEnc.statesMask)
    
    
    def check_ceu(self, fsm, explanation, agents, phi, psi):
        # For all states, check that
        # it is in phi | psi,
        # if it's in phi, it has a least one successor
        # if it's in psi, it has no successor
        extract = {explanation}
        states = set()
        while len(extract) > 0:
            expl = extract.pop()
            states.add(expl)
            self.assertTrue(expl.state <= phi | psi)
            if expl.state <= psi:
                self.assertTrue(len(expl.successors) == 0)
            else:
                self.assertTrue(len(expl.successors) > 0)
            for action, succ in expl.successors:
                if succ not in states:
                    extract.add(succ)
                
    
    def check_ceg(self, fsm, explanation, agents, phi):
        # For all states, check that
        # it is in phi,
        # it has at least one successor
        extract = {explanation}
        states = set()
        while len(extract) > 0:
            expl = extract.pop()
            states.add(expl)
            self.assertTrue(expl.state <= phi)
            self.assertTrue(len(expl.successors) > 0)
            for action, succ in expl.successors:
                if succ not in states:
                    extract.add(succ)
    
    
    def test_transmission_cex(self):
        fsm = self.transmission()
        
        spec = parseATL("<'transmitter'>X ~'received'")[0]
        agents = {atom.value for atom in spec.group}
        phi = evalATL(fsm, spec.child)
        self.assertTrue(check(fsm, spec))
        sat = evalATL(fsm, spec)
        initsat = sat & fsm.init
        first = fsm.pick_one_state(initsat)
        explanation = explain_cex(fsm, first, agents, phi)
        #self.show_cex(explanation, spec)
        self.check_cex(fsm, explanation, agents, phi)
    
    
    def test_cardgame_cex(self):
        fsm = self.cardgame()
        
        spec = parseATL("<'dealer'>X 'pcard = Ac'")[0]
        agents = {atom.value for atom in spec.group}
        phi = evalATL(fsm, spec.child)
        self.assertTrue(check(fsm, spec))
        sat = evalATL(fsm, spec)
        initsat = sat & fsm.init
        first = fsm.pick_one_state(initsat)
        explanation = explain_cex(fsm, first, agents, evalATL(fsm, spec.child))
        #self.show_cex(explanation, spec)
        self.check_cex(fsm, explanation, agents, phi)
        
        spec = parseATL("<'player'>X 'step = 1'")[0]
        agents = {atom.value for atom in spec.group}
        phi = evalATL(fsm, spec.child)
        self.assertTrue(check(fsm, spec))
        sat = evalATL(fsm, spec)
        initsat = sat & fsm.init
        first = fsm.pick_one_state(initsat)
        explanation = explain_cex(fsm, first, agents, evalATL(fsm, spec.child))
        #self.show_cex(explanation, spec)
        self.check_cex(fsm, explanation, agents, phi)
        

    def test_tictactoe_cex(self):
        fsm = self.tictactoe()
        
        spec = parseATL("'run = circle' -> <'circlep'>X 'run = cross'")[0]
        
        agents = {atom.value for atom in spec.right.group}
        phi = evalATL(fsm, spec.right.child)
        self.assertTrue(check(fsm, spec))
        sat = evalATL(fsm, spec)
        initsat = sat & fsm.init
        first = fsm.pick_one_state(initsat)
        explanation = explain_cex(fsm, first, agents, phi)
        #self.show_cex(explanation, spec.right)
        self.check_cex(fsm, explanation, agents, phi)
        
        spec = parseATL("'run = circle' -> <'circlep'>X 'board[1] = circle'")[0]
        agents = {atom.value for atom in spec.right.group}
        phi = evalATL(fsm, spec.right.child)
        self.assertTrue(check(fsm, spec))
        sat = evalATL(fsm, spec)
        initsat = sat & fsm.init
        first = fsm.pick_one_state(initsat)
        explanation = explain_cex(fsm, first, agents, phi)
        #self.show_cex(explanation, spec.right)
        self.check_cex(fsm, explanation, agents, phi)
        

    def test_transmission_cax(self):
        fsm = self.transmission()
        
        spec = parseATL("['sender']X ~'received'")[0]
        agents = {atom.value for atom in spec.group}
        phi = evalATL(fsm, spec.child)
        self.assertTrue(check(fsm, spec))
        sat = evalATL(fsm, spec)
        initsat = sat & fsm.init
        first = fsm.pick_one_state(initsat)
        explanation = explain_cax(fsm, first, agents, phi)
        #self.show_cex(explanation, spec)
        self.check_cax(fsm, explanation, agents, phi)
        
    
    def test_cardgame_cax(self):
        fsm = self.cardgame()
        
        spec = parseATL("['player']X 'pcard = Ac'")[0]
        agents = {atom.value for atom in spec.group}
        phi = evalATL(fsm, spec.child)
        self.assertTrue(check(fsm, spec))
        sat = evalATL(fsm, spec)
        initsat = sat & fsm.init
        first = fsm.pick_one_state(initsat)
        explanation = explain_cax(fsm, first, agents, evalATL(fsm, spec.child))
        #self.show_cex(explanation, spec)
        self.check_cax(fsm, explanation, agents, phi)
        
        spec = parseATL("['dealer']X 'win'")[0]
        agents = {atom.value for atom in spec.group}
        phi = evalATL(fsm, spec.child)
        # False since we do not want initial states
        #self.assertTrue(check(fsm, spec))
        sat = evalATL(fsm, spec)
        initsat = (sat & eval_simple_expression(fsm, 'step = 1') &
                   fsm.reachable_states)
        first = fsm.pick_one_state(initsat)
        explanation = explain_cax(fsm, first, agents, evalATL(fsm, spec.child))
        #self.show_cex(explanation, spec)
        self.check_cax(fsm, explanation, agents, phi)
        

    def test_tictactoe_cax(self):
        fsm = self.tictactoe()
        
        spec = parseATL("'run = circle' -> ['circlep']X 'run = cross'")[0]
        
        agents = {atom.value for atom in spec.right.group}
        phi = evalATL(fsm, spec.right.child)
        self.assertTrue(check(fsm, spec))
        sat = evalATL(fsm, spec)
        initsat = sat & fsm.init
        first = fsm.pick_one_state(initsat)
        explanation = explain_cax(fsm, first, agents, phi)
        #self.show_cex(explanation, spec.right)
        self.check_cax(fsm, explanation, agents, phi)
        
        spec = parseATL("'run = circle' -> ['crossp']X 'board[1] = circle'")[0]
        agents = {atom.value for atom in spec.right.group}
        phi = evalATL(fsm, spec.right.child)
        self.assertTrue(check(fsm, spec))
        sat = evalATL(fsm, spec)
        initsat = sat & fsm.init
        first = fsm.pick_one_state(initsat)
        explanation = explain_cax(fsm, first, agents, phi)
        #self.show_cex(explanation, spec.right)
        self.check_cax(fsm, explanation, agents, phi)
    
    
    def test_transmission_ceu(self):
        fsm = self.transmission()
        
        spec = parseATL("<'sender','transmitter'>['TRUE' U 'received']")[0]
        agents = {atom.value for atom in spec.group}
        phi = evalATL(fsm, spec.left)
        psi = evalATL(fsm, spec.right)
        self.assertTrue(check(fsm, spec))
        sat = evalATL(fsm, spec)
        initsat = sat & fsm.init
        first = fsm.pick_one_state(initsat)
        explanation = explain_ceu(fsm, first, agents, phi, psi)
        self.check_ceu(fsm, explanation, agents, phi, psi)
    
    
    def test_cardgame_ceu(self):
        fsm = self.cardgame()
        
        spec = parseATL("<'player'>['TRUE' U 'win']")[0]
        agents = {atom.value for atom in spec.group}
        phi = evalATL(fsm, spec.left)
        psi = evalATL(fsm, spec.right)
        self.assertTrue(check(fsm, spec))
        sat = evalATL(fsm, spec)
        initsat = sat & fsm.init
        first = fsm.pick_one_state(initsat)
        explanation = explain_ceu(fsm, first, agents, phi, psi)
        self.check_ceu(fsm, explanation, agents, phi, psi)
        
    
    @unittest.skip("Test too long")
    def test_tictactoe_ceu(self):
        fsm = self.tictactoe()
        
        spec = parseATL("<'circlep'> ['TRUE' U  'winner != circle' & 'run = stop']")[0]
        agents = {atom.value for atom in spec.group}
        phi = evalATL(fsm, spec.left)
        psi = evalATL(fsm, spec.right)
        self.assertTrue(check(fsm, spec))
        sat = evalATL(fsm, spec)
        initsat = sat & fsm.init
        first = fsm.pick_one_state(initsat)
        explanation = explain_ceu(fsm, first, agents, phi, psi)
        self.check_ceu(fsm, explanation, agents, phi, psi)
        
        spec = parseATL("<'circlep', 'crossp'> ['TRUE' U 'winner = empty' & 'run = stop']")[0]
        agents = {atom.value for atom in spec.group}
        phi = evalATL(fsm, spec.left)
        psi = evalATL(fsm, spec.right)
        self.assertTrue(check(fsm, spec))
        sat = evalATL(fsm, spec)
        initsat = sat & fsm.init
        first = fsm.pick_one_state(initsat)
        explanation = explain_ceu(fsm, first, agents, phi, psi)
        self.check_ceu(fsm, explanation, agents, phi, psi)
        
    
    def test_transmission_cau(self):
        fsm = self.transmission()
        
        spec = parseATL("['transmitter']['TRUE' U ~'received']")[0]
        agents = {atom.value for atom in spec.group}
        phi = evalATL(fsm, spec.left)
        psi = evalATL(fsm, spec.right)
        self.assertTrue(check(fsm, spec))
        sat = evalATL(fsm, spec)
        initsat = sat & fsm.init
        first = fsm.pick_one_state(initsat)
        explanation = explain_cau(fsm, first, agents, phi, psi)
        self.check_ceu(fsm, explanation, agents, phi, psi)
    
    
    def test_cardgame_cau(self):
        fsm = self.cardgame()
        
        spec = parseATL("['dealer']['TRUE' U 'win']")[0]
        agents = {atom.value for atom in spec.group}
        phi = evalATL(fsm, spec.left)
        psi = evalATL(fsm, spec.right)
        self.assertTrue(check(fsm, spec))
        sat = evalATL(fsm, spec)
        initsat = sat & fsm.init
        first = fsm.pick_one_state(initsat)
        explanation = explain_cau(fsm, first, agents, phi, psi)
        self.check_ceu(fsm, explanation, agents, phi, psi)
        
        spec = parseATL("['dealer']['TRUE' U 'lose']")[0]
        agents = {atom.value for atom in spec.group}
        phi = evalATL(fsm, spec.left)
        psi = evalATL(fsm, spec.right)
        self.assertTrue(check(fsm, spec))
        sat = evalATL(fsm, spec)
        initsat = sat & fsm.init
        first = fsm.pick_one_state(initsat)
        explanation = explain_cau(fsm, first, agents, phi, psi)
        self.check_ceu(fsm, explanation, agents, phi, psi)
        
    
    @unittest.skip("Test too long")
    def test_tictactoe_cau(self):
        fsm = self.tictactoe()
        
        spec = parseATL("['circlep']['TRUE' U  ('winner != circle' & 'run = stop')]")[0]
        agents = {atom.value for atom in spec.group}
        phi = evalATL(fsm, spec.left)
        psi = evalATL(fsm, spec.right)
        self.assertTrue(check(fsm, spec))
        sat = evalATL(fsm, spec)
        
        self.assertEqual(sat, fp(lambda Y : psi | (phi & fsm.pre_nstrat(Y, agents)), BDD.false(fsm.bddEnc.DDmanager)))
        
        initsat = sat & fsm.init
        first = fsm.pick_one_state(initsat)
        explanation = explain_cau(fsm, first, agents, phi, psi)
        self.check_ceu(fsm, explanation, agents, phi, psi)

        
    def test_transmission_ceg(self):
        fsm = self.transmission()
        
        spec = parseATL("<'transmitter'> G ~'received'")[0]
        agents = {atom.value for atom in spec.group}
        phi = evalATL(fsm, spec.child)
        self.assertTrue(check(fsm, spec))
        sat = evalATL(fsm, spec)
        initsat = sat & fsm.init
        first = fsm.pick_one_state(initsat)
        explanation = explain_ceg(fsm, first, agents, phi)
        self.check_ceg(fsm, explanation, agents, phi)
        
        
        spec = parseATL("<'sender'> G ~'received'")[0]
        agents = {atom.value for atom in spec.group}
        phi = evalATL(fsm, spec.child)
        self.assertTrue(check(fsm, spec))
        sat = evalATL(fsm, spec)
        initsat = sat & fsm.init
        first = fsm.pick_one_state(initsat)
        explanation = explain_ceg(fsm, first, agents, phi)
        self.check_ceg(fsm, explanation, agents, phi)
    
    
    def test_cardgame_ceg(self):
        fsm = self.cardgame()
        
        spec = parseATL("<'player'> G ~'win'")[0]
        agents = {atom.value for atom in spec.group}
        phi = evalATL(fsm, spec.child)
        self.assertTrue(check(fsm, spec))
        sat = evalATL(fsm, spec)
        initsat = sat & fsm.init
        first = fsm.pick_one_state(initsat)
        explanation = explain_ceg(fsm, first, agents, phi)
        self.check_ceg(fsm, explanation, agents, phi)
        
        spec = parseATL("<'player','dealer'> G ~'win'")[0]
        agents = {atom.value for atom in spec.group}
        phi = evalATL(fsm, spec.child)
        self.assertTrue(check(fsm, spec))
        sat = evalATL(fsm, spec)
        initsat = sat & fsm.init
        first = fsm.pick_one_state(initsat)
        explanation = explain_ceg(fsm, first, agents, phi)
        self.check_ceg(fsm, explanation, agents, phi)
        
    
    @unittest.skip("Test too long.")
    def test_tictactoe_ceg(self):
        fsm = self.tictactoe()
        
        spec = parseATL("<'circlep', 'crossp'> G 'winner = empty'")[0]
        agents = {atom.value for atom in spec.group}
        phi = evalATL(fsm, spec.child)
        self.assertTrue(check(fsm, spec))
        sat = evalATL(fsm, spec)
        initsat = sat & fsm.init
        first = fsm.pick_one_state(initsat)
        explanation = explain_ceg(fsm, first, agents, phi)
        self.check_ceg(fsm, explanation, agents, phi)
        
        spec = parseATL("<'circlep'> G 'winner != circle'")[0]
        agents = {atom.value for atom in spec.group}
        phi = evalATL(fsm, spec.child)
        self.assertTrue(check(fsm, spec))
        sat = evalATL(fsm, spec)
        initsat = sat & fsm.init
        first = fsm.pick_one_state(initsat)
        explanation = explain_ceg(fsm, first, agents, phi)
        self.check_ceg(fsm, explanation, agents, phi)
        
    
    def test_transmission_cag(self):
        fsm = self.transmission()
        
        spec = parseATL("['sender'] G ~'received'")[0]
        agents = {atom.value for atom in spec.group}
        phi = evalATL(fsm, spec.child)
        self.assertTrue(check(fsm, spec))
        sat = evalATL(fsm, spec)
        initsat = sat & fsm.init
        first = fsm.pick_one_state(initsat)
        explanation = explain_cag(fsm, first, agents, phi)
        self.check_ceg(fsm, explanation, agents, phi)
        
        
        spec = parseATL("['transmitter'] G ~'received'")[0]
        agents = {atom.value for atom in spec.group}
        phi = evalATL(fsm, spec.child)
        self.assertTrue(check(fsm, spec))
        sat = evalATL(fsm, spec)
        initsat = sat & fsm.init
        first = fsm.pick_one_state(initsat)
        explanation = explain_cag(fsm, first, agents, phi)
        self.check_ceg(fsm, explanation, agents, phi)
    
    
    def test_cardgame_cag(self):
        fsm = self.cardgame()
        
        spec = parseATL("['dealer'] G ~'win'")[0]
        agents = {atom.value for atom in spec.group}
        phi = evalATL(fsm, spec.child)
        self.assertTrue(check(fsm, spec))
        sat = evalATL(fsm, spec)
        initsat = sat & fsm.init
        first = fsm.pick_one_state(initsat)
        explanation = explain_cag(fsm, first, agents, phi)
        self.check_ceg(fsm, explanation, agents, phi)
        
    
    @unittest.skip("Test too long.")
    def test_tictactoe_cag(self):
        fsm = self.tictactoe()
        
        spec = parseATL("['circlep'] G 'winner != circle'")[0]
        agents = {atom.value for atom in spec.group}
        phi = evalATL(fsm, spec.child)
        self.assertTrue(check(fsm, spec))
        sat = evalATL(fsm, spec)
        initsat = sat & fsm.init
        first = fsm.pick_one_state(initsat)
        explanation = explain_cag(fsm, first, agents, phi)
        self.check_ceg(fsm, explanation, agents, phi)
    
    
    def test_cardgame_cew(self):
        fsm = self.cardgame()
        
        spec = parseATL("<'player'>[~'lose' W 'win']")[0]
        agents = {atom.value for atom in spec.group}
        phi = evalATL(fsm, spec.left)
        psi = evalATL(fsm, spec.right)
        self.assertTrue(check(fsm, spec))
        sat = evalATL(fsm, spec)
        initsat = sat & fsm.init
        first = fsm.pick_one_state(initsat)
        explanation = explain_cew(fsm, first, agents, phi, psi)
        self.check_ceu(fsm, explanation, agents, phi, psi)
        
        
        spec = parseATL("<'player'>[~'lose' W 'step = 3']")[0]
        agents = {atom.value for atom in spec.group}
        phi = evalATL(fsm, spec.left)
        psi = evalATL(fsm, spec.right)
        self.assertTrue(check(fsm, spec))
        sat = evalATL(fsm, spec)
        initsat = sat & fsm.init
        first = fsm.pick_one_state(initsat)
        explanation = explain_cew(fsm, first, agents, phi, psi)
        self.check_ceu(fsm, explanation, agents, phi, psi)
        
        
    def test_cardgame_caw(self):
        fsm = self.cardgame()
        
        spec = parseATL("['dealer'][~'lose' W 'win']")[0]
        agents = {atom.value for atom in spec.group}
        phi = evalATL(fsm, spec.left)
        psi = evalATL(fsm, spec.right)
        self.assertTrue(check(fsm, spec))
        sat = evalATL(fsm, spec)
        initsat = sat & fsm.init
        first = fsm.pick_one_state(initsat)
        explanation = explain_caw(fsm, first, agents, phi, psi)
        self.check_ceu(fsm, explanation, agents, phi, psi)
        
        
        spec = parseATL("['dealer'][~'lose' W 'step = 3']")[0]
        agents = {atom.value for atom in spec.group}
        phi = evalATL(fsm, spec.left)
        psi = evalATL(fsm, spec.right)
        self.assertTrue(check(fsm, spec))
        sat = evalATL(fsm, spec)
        initsat = sat & fsm.init
        first = fsm.pick_one_state(initsat)
        explanation = explain_caw(fsm, first, agents, phi, psi)
        self.check_ceu(fsm, explanation, agents, phi, psi)


from ..tlace.tlace import (Tlacenode, Tlacebranch, EpistemicBranch,
                           TemporalBranch)

class PartialTlacenode(Tlacenode):
    """
    A PartialTlacenode is a TLACE node with some potentially unexplained
    branches. This means that some branch may not be expandable and must
    be explained before. Its branches can be TLACE branches (instances of
    (a subclass of) the TlaceBranch class) or specifications, the former
    being an explained branch while the latter representing an unexplained
    branch.
    """
    
    def __init__(self, state, atomics=None, branches=None, universals=None):
        """
        Create a new partial TLACE node.
        
        state -- a State representing the state of the TLACE node
        atomics -- a list of atomic propositions,
                   represented by CTLK AST instances
        branches -- a list of TLACE branches represented by Tlacebranches
        universals -- a list of universal formulas,
                      represented by CTLK AST instances
        """
        self._state = state
        self._atomics = atomics or []
        self._branches = branches or []
        self._universals = universals or []


import unittest

from pynusmv.fsm import BddFsm
from pynusmv.init import init_nusmv, deinit_nusmv

from pynusmv_tools.arctl.parsing import parseArctl
from pynusmv_tools.arctl.check import checkArctl
from pynusmv_tools.arctl.tlace.explain import explain_witness, explain_countex

from pynusmv_tools.arctl.tlace.xml import xml_representation


class TestXML(unittest.TestCase):
    
    def setUp(self):
        init_nusmv()
    
    def tearDown(self):
        deinit_nusmv()
    
    
    def init_model(self):
        fsm = BddFsm.from_filename("tests/pynusmv_tools/arctl/counters.smv")
        self.assertIsNotNone(fsm)
        return fsm
        
    
    def init_finite_model(self):
        fsm = BddFsm.from_filename("tests/pynusmv_tools/arctl/finite_counters.smv")
        self.assertIsNotNone(fsm)
        return fsm
        

    def test_full(self):
        fsm = self.init_model()
        
        spec = parseArctl("A<'run = rc1'>["
                            "E<'TRUE'>G 'c1.c < 3' W "
                            "A<'TRUE'>X 'c1.c = 0' ]")[0]
                            
        (result, expl) = checkArctl(fsm, spec, explain_witness, explain_countex)
        self.assertTrue(result)
        
        self.assertIsNotNone(xml_representation(fsm, expl, spec))
        print(xml_representation(fsm, expl, spec))
        
        
        
    def test_full_finite(self):
        fsm = self.init_finite_model()
        
        spec = parseArctl("A<'TRUE'>F (~ E<'TRUE'>X 'TRUE')")[0]
        
        (result, expl) = checkArctl(fsm, spec, explain_witness, explain_countex)
        self.assertTrue(result)
        
        self.assertIsNotNone(xml_representation(fsm, expl, spec))
        print(xml_representation(fsm, expl, spec))


import sys

from pynusmv.prop import PropDb
from pynusmv.prop import propTypes
from pynusmv.init import init_nusmv, deinit_nusmv
from pynusmv import glob
from pynusmv.fsm import BddFsm
from pynusmv.exception import PyNuSMVError

from .eval import eval_ctl

def check(modelPath, evalSpecs=True):
    init_nusmv()
    # Initialize the model
    fsm = BddFsm.from_filename(modelPath)
    propDb = glob.prop_database()

    # Check all CTL properties
    if evalSpecs:
        for prop in propDb:
            #  Check type
            if prop.type == propTypes['CTL']:
                spec = prop.exprcore

                # Get violating states
                violating = (fsm.init & ~eval_ctl(fsm, spec) &
                             fsm.state_constraints)
                print('Specification',str(spec), 'is',
                      str(violating.is_false()))
                # We could generate counter-examples here
    deinit_nusmv()

def main():
    if len(sys.argv) < 2:
        print("[ERROR] Missing model.")
    else:
        try:
            check(sys.argv[-1], not "-dcx" in sys.argv)
        except PyNuSMVError as e:
            print("[Error]", str(e))

if __name__ == '__main__':
    main()



import unittest

from pynusmv_lower_interface.nusmv.cmd import cmd

from pynusmv.fsm import BddFsm
from pynusmv.prop import PropDb
from pynusmv import glob

from pynusmv_tools.tlace.check import check

from pynusmv.init import init_nusmv, deinit_nusmv

class TestTlaceGeneration(unittest.TestCase):
    
    def setUp(self):
        init_nusmv()
        
    def tearDown(self):
        deinit_nusmv()
    
    
    def test_check_violated_spec(self):
        # Initialize the model
        ret = cmd.Cmd_SecureCommandExecute("read_model -i " 
                                     "tests/pynusmv_tools/tlace/admin.smv")
        self.assertEqual(ret, 0, "cannot read the model")
        ret = cmd.Cmd_SecureCommandExecute("go")
        self.assertEqual(ret, 0, "cannot build the model")
        
        propDb = glob.prop_database()
        master = propDb.master
        fsm = propDb.master.bddFsm
        self.assertTrue(propDb.get_size() >= 1, "propDb has no properties")
        prop = propDb.get_prop_at_index(0)
        spec = prop.exprcore
        
        res = check(fsm, spec)
        self.assertFalse(res[0], "spec should be violated")
        self.assertIsNotNone(res[1], "TLACE should be given")
        
    
    def test_check_violated_spec2(self):
        # Initialize the model
        ret = cmd.Cmd_SecureCommandExecute("read_model -i"
                                     "tests/pynusmv_tools/tlace/admin.smv")
        self.assertEqual(ret, 0, "cannot read the model")
        ret = cmd.Cmd_SecureCommandExecute("go")
        self.assertEqual(ret, 0, "cannot build the model")
        
        propDb = glob.prop_database()
        master = propDb.master
        fsm = propDb.master.bddFsm
        self.assertTrue(propDb.get_size() >= 2, "propDb misses some props")
        prop = propDb.get_prop_at_index(1)
        self.assertIsNotNone(prop, "prop should not be None")
        spec = prop.exprcore
        
        res = check(fsm, spec)
        self.assertFalse(res[0], "spec should be violated")
        self.assertIsNotNone(res[1], "TLACE should be given")
        
    
    def test_check_satisfied_spec(self):
        # Initialize the model
        ret = cmd.Cmd_SecureCommandExecute("read_model -i"
                                     "tests/pynusmv_tools/tlace/admin.smv")
        self.assertEqual(ret, 0, "cannot read the model")
        ret = cmd.Cmd_SecureCommandExecute("go")
        self.assertEqual(ret, 0, "cannot build the model")
        
        propDb = glob.prop_database()
        master = propDb.master
        fsm = propDb.master.bddFsm
        self.assertTrue(propDb.get_size() >= 3, "propDb misses some props")
        prop = propDb.get_prop_at_index(2)
        self.assertIsNotNone(prop, "prop should not be None")
        spec = prop.exprcore
        
        res = check(fsm, spec)
        self.assertTrue(res[0], "spec should be satisfied")
        self.assertIsNone(res[1], "TLACE should not exist")
        
    
    def test_check_ex_spec(self):
        # Initialize the model
        ret = cmd.Cmd_SecureCommandExecute("read_model -i"
                                     "tests/pynusmv_tools/tlace/admin.smv")
        self.assertEqual(ret, 0, "cannot read the model")
        ret = cmd.Cmd_SecureCommandExecute("go")
        self.assertEqual(ret, 0, "cannot build the model")
        
        propDb = glob.prop_database()
        master = propDb.master
        fsm = propDb.master.bddFsm
        self.assertTrue(propDb.get_size() >= 4, "propDb misses some props")
        prop = propDb.get_prop_at_index(3)
        self.assertIsNotNone(prop, "prop should not be None")
        spec = prop.exprcore
        
        res = check(fsm, spec)
        self.assertFalse(res[0], "spec should be violated")
        self.assertIsNotNone(res[1], "TLACE should be given")


import unittest

from pynusmv.fsm import BddFsm
from pynusmv.dd import BDD
from pynusmv.init import init_nusmv, deinit_nusmv

from pynusmv_tools.arctl.parsing import parseArctl
from pynusmv_tools.arctl.ast import (Atom, Not, And, Or, Implies, Iff,
                             EaX, EaG, EaF, EaU, EaW,
                             AaX, AaG, AaF, AaU, AaW)
from pynusmv_tools.arctl.eval import evalArctl, evalArctl_from_string as evalStr


class TestEval(unittest.TestCase):
    
    def setUp(self):
        init_nusmv()
    
    def tearDown(self):
        deinit_nusmv()
    
    
    def init_model(self):
        fsm = BddFsm.from_filename("tests/pynusmv_tools/arctl/model.smv")
        self.assertIsNotNone(fsm)
        return fsm
        
    
    def init_finite_model(self):
        fsm = BddFsm.from_filename("tests/pynusmv_tools/arctl/finite_model.smv")
        self.assertIsNotNone(fsm)
        return fsm
        

    def test_atom(self):
        fsm = self.init_model()
        
        specs = parseArctl("'c'")
        self.assertEqual(len(specs), 1)
        spec = specs[0]
        self.assertEqual(type(spec), Atom)
        self.assertEqual(spec.value, "c")
        
        c = evalArctl(fsm, spec)
        self.assertIsNotNone(c)
        self.assertTrue(fsm.init <= c)
        
        
    def test_not(self):
        fsm = self.init_model()
        
        specs = parseArctl("~'c'")
        self.assertEqual(len(specs), 1)
        spec = specs[0]
        
        nc = evalArctl(fsm, spec)
        self.assertTrue((nc & fsm.init).is_false())
        
        
    def test_and(self):
        fsm = self.init_model()
        
        candi = evalStr(fsm, "('c' & 'i')")
        c = evalStr(fsm, "'c'")
        i = evalStr(fsm, "'i'")
        self.assertEqual(candi, c & i)
        
        
    def test_or(self):
        fsm = self.init_model()
        
        cordi = evalStr(fsm, "('c' | 'i')")
        c = evalStr(fsm, "'c'")
        i = evalStr(fsm, "'i'")
        self.assertEqual(cordi, c | i)
        
        
    def test_implies(self):
        fsm = self.init_model()
        
        cimpli = evalStr(fsm, "('c' -> 'i')")
        c = evalStr(fsm, "'c'")
        i = evalStr(fsm, "'i'")
        self.assertTrue(~c <= cimpli)
        self.assertTrue(i <= cimpli)
        self.assertEqual(~c | i, cimpli)
        
        
    def test_iff(self):
        fsm = self.init_model()
        
        ciffi = evalStr(fsm, "('c' <-> 'i')")
        c = evalStr(fsm, "'c'")
        i = evalStr(fsm, "'i'")
        self.assertTrue(c & i <= ciffi)
        self.assertTrue(~c & ~i <= ciffi)
        self.assertEqual((c & i) | (~c & ~i), ciffi)
        
        
    def test_eax(self):
        fsm = self.init_model()
        
        eaaxi = evalStr(fsm, "E<'a'>X 'i'")
        self.assertTrue((fsm.init & eaaxi).is_false())
        eaaxnc = evalStr(fsm, "E<'a'>X ~'c'")
        c = evalStr(fsm, "'c'")
        i = evalStr(fsm, "'i'")
        self.assertTrue(eaaxnc.isnot_false())
        self.assertEqual(eaaxnc, c.iff(i))
        
    
    def test_neaxt(self):
        fsm = self.init_finite_model()
        
        candi = evalStr(fsm, "('c' & ~'i')")
        neaxt = evalStr(fsm, "~E<'a'>X 'TRUE'")
        
        self.assertTrue(candi <= neaxt)
        
        
    def test_aax(self):
        fsm = self.init_model()
        aaxnc = evalStr(fsm, "A<'a'>X ~'c'")
        c = evalStr(fsm, "'c'")
        i = evalStr(fsm, "'i'")
        self.assertEqual(aaxnc, c.iff(i))
    
      
    def test_aax_finite(self):
        fsm = self.init_finite_model()
        aaxnc = evalStr(fsm, "A<'a'>X ~'c'")
        c = evalStr(fsm, "'c'")
        i = evalStr(fsm, "'i'")
        self.assertEqual(fsm.init, aaxnc)
        self.assertEqual(aaxnc, c & i)
        
    
    def test_eaf(self):
        fsm = self.init_model()
        eafnc = evalStr(fsm, "E<'a'>F ~'c'")
        c = evalStr(fsm, "'c'")
        i = evalStr(fsm, "'i'")
        self.assertEqual(eafnc, ~(c & ~i))
        
        
    def test_eaf_finite(self):
        fsm = self.init_finite_model()
        eafnc = evalStr(fsm, "E<'a'>F ~'c'")
        c = evalStr(fsm, "'c'")
        i = evalStr(fsm, "'i'")
        self.assertTrue(c.iff(i) <= eafnc)
      
  
    def test_aaf(self):
        fsm = self.init_model()
        aafni = evalStr(fsm, "A<'a'>F ~'i'")
        true = BDD.true(fsm.bddEnc.DDmanager)
        self.assertEqual(aafni, true)
            
    
    def test_aaf_finite(self):
        fsm = self.init_finite_model()
        aafni = evalStr(fsm, "A<'a'>F ~'i'")
        c, i = evalStr(fsm, "'c'"), evalStr(fsm, "'i'")
        
        self.assertEqual(aafni, c | ~i)
        
    
    def test_eag(self):
        fsm = self.init_model()
        enagc = evalStr(fsm, "E<~'a'>G 'c'")
        c = evalStr(fsm, "'c'")
        i = evalStr(fsm, "'i'")
        
        self.assertTrue(fsm.init <= enagc)
        self.assertEqual(enagc, c)
        
    
    def test_eag_finite(self):
        fsm = self.init_finite_model()
        enagc = evalStr(fsm, "E<~'a'>G 'c'")
        c, i = evalStr(fsm, "'c'"), evalStr(fsm, "'i'")
        
        self.assertEqual(enagc, c)
        
        
    def test_aag(self):
        fsm = self.init_model()
        c, i = evalStr(fsm, "'c'"), evalStr(fsm, "'i'")

        anagc = evalStr(fsm, "A<~'a'>G 'c'")
        self.assertEqual(anagc, c)
        
        atgc = evalStr(fsm, "A<'TRUE'>G 'c'")
        self.assertEqual(atgc, c & ~i)
        
        
    def test_aag_finite(self):
        fsm = self.init_finite_model()
        c, i = evalStr(fsm, "'c'"), evalStr(fsm, "'i'")
        anagc = evalStr(fsm, "A<~'a'>G 'c'")
        self.assertEqual(anagc, c)
        
        
    def test_eau(self):
        fsm = self.init_model()
        c, i = evalStr(fsm, "'c'"), evalStr(fsm, "'i'")
        eacui = evalStr(fsm, "E<'a'>['c' U 'i']")
        self.assertEqual(fsm.init | (~c & i), eacui)
        
        
    def test_eau_finite(self):
        fsm = self.init_finite_model()
        c, i = evalStr(fsm, "'c'"), evalStr(fsm, "'i'")
        enacuni = evalStr(fsm, "E<~'a'>['c'U~'i']")
        self.assertEqual(enacuni, c | ~i)
        
        
    def test_aau(self):
        fsm = self.init_model()
        c, i = evalStr(fsm, "'c'"), evalStr(fsm, "'i'")
        aacuni = evalStr(fsm, "A<'a'>['c'U~'i']")
        self.assertEqual(aacuni, c | ~i)
        
        
    def test_aau_finite(self):
        fsm = self.init_finite_model()
        c, i = evalStr(fsm, "'c'"), evalStr(fsm, "'i'")
        aacuni = evalStr(fsm, "A<~'a'>['c'U~'i']")
        self.assertEqual(aacuni, c | ~i)
        
        
    def test_eaw(self):
        fsm = self.init_model()
        c, i = evalStr(fsm, "'c'"), evalStr(fsm, "'i'")
        eacui = evalStr(fsm, "E<'a'>['c' W 'i']")
        self.assertEqual(c | i, eacui)
        
        
    def test_eaw_finite(self):
        fsm = self.init_finite_model()
        c, i = evalStr(fsm, "'c'"), evalStr(fsm, "'i'")
        enacuni = evalStr(fsm, "E<~'a'>['c'W~'i']")
        self.assertEqual(enacuni, c | ~i)
        
        
    def test_aaw(self):
        fsm = self.init_model()
        c, i = evalStr(fsm, "'c'"), evalStr(fsm, "'i'")
        aacuni = evalStr(fsm, "A<'a'>['c'W~'i']")
        self.assertEqual(aacuni, c | ~i)
        
        
    def test_aaw_finite(self):
        fsm = self.init_finite_model()
        c, i = evalStr(fsm, "'c'"), evalStr(fsm, "'i'")
        aacuni = evalStr(fsm, "A<~'a'>['c'W~'i']")
        self.assertEqual(aacuni, c | ~i)


import sys
import argparse
from colorsys import rgb_to_hsv

from pynusmv.init import init_nusmv, deinit_nusmv
from pynusmv import glob
from pynusmv.exception import PyNuSMVError

# Choices of representation
# -------------------------
# Initial states are marked with a bold border.
#
# Fair states and actions and colored differently according to the fairness
# constraint.
# FIXME Graphviz does not recognize gradient colors

__fair_attr = None
def fairness_attr(fsm):
    """
    Return a dictionary of fairness constraint -> DOT attributes to use in the
    DOT representation of a model to indicate the different fairness
    constraints.
    
    fsm -- the BddFsm of the model.
    
    """
    global __fair_attr
    if not __fair_attr:
        __fair_attr = dict()
        nbfair = len(fsm.fairness_constraints)
        curid = 0
        for f in fsm.fairness_constraints:
            __fair_attr[f] = " ".join(str(val) for val in 
                                  rgb_to_hsv(curid/nbfair, 1, 1 - curid/nbfair))
            curid += 1
        
    return __fair_attr
    

def node(fsm, state, ids):
    """
    Return the DOT representation of the given state.
    
    fsm -- the BddFsm of the model
    state -- a State
    ids -- a dictionary of the ids of some states. Must contain state.
    
    """
    attr = set()
    
    # Label
    attr.add("label=\"" + '\\n'.join(var+"="+val for var, val
                                     in sorted(state.
                                               get_str_values().items()))
                        + "\"")
    
    # Initial state
    if state <= fsm.init:
        attr.add("penwidth=5")
        
    # Fairness
    fair_styles = set()
    fair_attr = fairness_attr(fsm)
    for f in fsm.fairness_constraints:
        if state <= f:
            fair_styles.add(fair_attr[f])
    if len(fair_styles):
        attr.add("style=filled")
        attr.add("fillcolor=\"" + ":".join(fair_styles) + "\"")
        
    if len(attr):
        attrstr = "[" + ','.join(attr) + "]"
    else:
        attrstr = ""
    return ids[state] + " " + attrstr + ";"
    
    
def edge(fsm, source, inputs, target, ids):
    """
    Return the DOT representation of the given transition.
    
    fsm -- the BddFsm of the model
    source -- the source of the transition
    inputs -- the action of the transition, or None
    target -- the target of the transition
    ids -- a dictionary of the ids of some states.
           Must contain source and target.
    
    """
    attr = set()
    
    if inputs is not None:
        # Label
        attr.add("label=\"" + "\\n".join(var+"="+val for var, val
                                        in sorted(inputs.
                                                  get_str_values().items()))
                            + "\"")
        
        # Fairness
        fair_styles = set()
        fair_attr = fairness_attr(fsm)
        for f in fsm.fairness_constraints:
            if inputs <= f:
                fair_styles.add(fair_attr[f])
        if len(fair_styles):
            attr.add("penwidth=2")
            attr.add("color=\"" + ":".join(fair_styles) + "\"")
    
    if len(attr):
        attrstr = "[" + ','.join(attr) + "]"
    else:
        attrstr = ""
    return ids[source] + "->" + ids[target] + " " + attrstr + ";"
    

def dumpDot(fsm):
    """
    Print a DOT representation of the reachable graph of fsm.
    
    fsm -- the BddFsm of the model.
    
    """
    
    init = fsm.init & fsm.state_constraints & fsm.bddEnc.statesMask
    fair_attr = fairness_attr(fsm)
    states = fsm.pick_all_states(fsm.reachable_states)
    transitions = {(s1,i,s2)
                   for s1 in states
                   for s2 in 
                   fsm.pick_all_states(fsm.post(s1) & fsm.reachable_states)
                   for i in (fsm.pick_all_inputs(
                                         fsm.get_inputs_between_states(s1,s2))
                             if len(fsm.bddEnc.inputsVars) > 0
                             else {None})
                   if s2 <= fsm.post(s1, inputs=i)}
    
    dot = ["digraph {"]
    
    # Add states
    # Generate ids
    ids = dict()
    curid = 1
    for state in states:
        ids[state] = "s" + str(curid)
        curid = curid + 1
    
    # Print states
    for state in states:
        dot.append(node(fsm, state, ids))
      
        
    # Add transitions
    for s1, i, s2 in transitions:
        dot.append(edge(fsm, s1, i, s2, ids))
    
    dot.append("}")
    
    return "\n".join(dot)
    
    
def process(allargs):
    """
    Process program arguments and dump the model.
    Write on standard output the DOT format of the dumped model.
    
    allargs -- a sys.args-like arguments list, without script name.
    
    """
    
    # Parse arguments
    parser = argparse.ArgumentParser(description='SMV model DOT dumper.')
    # Populate arguments: for now, only the model
    parser.add_argument('model', help='the SMV model')
    args = parser.parse_args(allargs)
    
    # Initialize the model
    glob.load_from_file(args.model)
    glob.compute_model()
    fsm = glob.prop_database().master.bddFsm
    
    try:
        print(dumpDot(fsm))
    except PyNuSMVError as e:
        print("[ERROR]", str(e), file=sys.stderr)


def main():
    with init_nusmv():
        process(sys.argv[1:])
        
if __name__ == '__main__':
    main()



"""
This module contains the functions to perform the bounded model checking of a 
given LTL property. 
"""
from pynusmv.bmc.glob   import master_be_fsm
from pynusmv.sat        import SatSolverResult, SatSolverFactory, Polarity
from pynusmv.bmc.utils  import generate_counter_example
from pynusmv_tools.bmcLTL.gen   import generate_problem

def check_ltl_onepb(fml, length, no_fairness=False, no_invar=False, dry_run=False):
    """
    This function verifies that the given FSM satisfies the given property
    for paths with an exact length of `length`.
    
    :param fml: an LTL formula parsed with `pynusmv_tools.bmcLTL.parsing` (hence the 
        abstract syntax tree of that formula). Note, this is *NOT* the NuSMV
        format (Node).
    :param length: the exact length of the considered paths
    :param no_fairness: a flag telling whether or not the generated problem should
        focus on fair executions only (the considered fairness constraints must
        be declared in the SMV model).
    :param no_invar: a flag telling whether or not the generated problem 
        should enforce the declared invariants (these must be declared in the
        SMV text).
    :return: a tuple ('OK', None) if the property is satisfied on all paths of 
        length `length`
    :return: a tuple ('Violation', counter_example) if the property is violated. 
        The counter_example passed along is a trace leading to a violation of
        the property
    """
    fsm    = master_be_fsm()
    pb     = generate_problem(fml, fsm, length, no_fairness, no_invar)
    
    if not dry_run:
        cnf    = pb.to_cnf(Polarity.POSITIVE)
        
        solver = SatSolverFactory.create()
        solver+= cnf
        solver.polarity(cnf, Polarity.POSITIVE)
        
        if solver.solve() == SatSolverResult.SATISFIABLE:
            cnt_ex = generate_counter_example(fsm, pb, solver, length, str(fml))
            return ("Violation", cnt_ex)
        else:
            return ("Ok", None)
    return ("Ok", None)
    
def check_ltl(fml, bound, no_fairness=False, no_invar=False, dry_run=False):
    """
    This function performs the bounded model checking of the formula given in 
    text format (as specified per the grammar in `parsing` module). It verifies
    that the property holds for all path lengths from 0 to bound.
    
    :param fml: an LTL formula parsed with `pynusmv_tools.bmcLTL.parsing` (hence the 
        abstract syntax tree of that formula). Note, this is *NOT* the NuSMV
        format (Node).
    :param bound: the maximum length of a path in the verification.
    :param no_fairness: a flag telling whether or not the generated problem should
        focus on fair executions only (the considered fairness constraints must
        be declared in the SMV model).
    :param no_invar: a flag telling whether or not the generated problem 
        should enforce the declared invariants (these must be declared in the
        SMV text).
    :return: a tuple (status, len, trace) where status is 'Ok', len = bound and
        trace is None when no counter example was identified. Otherwise, 
        status = 'Violation', len the number of steps to reach a violation and
        trace is a counter example leading to a property violation.
    """
    for i in range(bound+1):
        status, trace = check_ltl_onepb(fml, i, no_fairness, no_invar, dry_run) 
        if status != "Ok":
            return (status, i, trace)
        else:
            print("-- No problem at length {}".format(i))
            
    return ("Ok", bound, None)
    


"""
Indent module allows to indent a given string.
"""

import sys

__value = "   "
__count = 0

def reset():
    """Reset the indentation to 0"""
    global __count
    __count = 0
    
def inc():
    """Increase the indentation"""
    global __count
    __count += 1
    
def dec():
    """Decrease the indentation"""
    global __count
    __count -= 1
    
def indent(s):
    """Indent s with the current indentation"""
    global __value
    global __count
    return __value * __count + s


import sys
import argparse
from pynusmv.init import init_nusmv, deinit_nusmv
from ..mas import glob
from ..atlkFO.parsing import parseATLK
from .eval import evalATLK as evalATLK_naive
from .evalGen import evalATLK as evalATLK_gen
from .evalOpt import evalATLK as evalATLK_opt
from .evalMem import evalATLK as evalATLK_mem
from .evalSymb import evalATLK as evalATLK_symb
from .evalPartial import evalATLK as evalATLK_partial
from .evalGenSI import evalATLK as evalATLK_genSI
from .evalPartialSI import evalATLK as evalATLK_partialSI
from pyparsing import ParseException
from pynusmv.exception import PyNuSMVError

from . import config

__implementations = {"naive" : evalATLK_naive,
                     "generator" : evalATLK_gen,
                     "optimized" : evalATLK_opt,
                     "memory" : evalATLK_mem,
                     "partial" : evalATLK_partial,
                     "symbolic" : evalATLK_symb,
                     "generatorSI" : evalATLK_genSI,
                     "partialSI" : evalATLK_partialSI}

def check(mas, spec, variant="SF", semantics="group", implem="generator"):
    """
    Return whether the system satisfies the ATLK specification.
    
    mas -- the system
    spec -- the specification, as a string
    variant -- the variant of the algorithm to evaluate strategic operators;
               must be
               * "SF" for the standard way: splitting in uniform strategies then
                 filtering winning states,
               * "FS" for the alternating way: filtering winning states, then
                 splitting one conflicting equivalence class, then recurse
               * "FSF" for the filter-split-filter way: filtering winning states
                 then splitting all remaining actions into uniform strategies,
                 then filtering final winning states.
    implem -- the particular implementation to use for strategic operators
              evaluation:
              * "naive" the naive implementation;
              * "generator" the generator-based implementation;
              * "optimized" a memory-optimized version;
              * "memory" another memory-optimized version;
              * "partial" a version based on partial strategies;
              * "symbolic" a version based on a fully symbolic approach.
    semantics -- the semantics to use for starting point and strategy point
                 equivalence:
                 * "group" for the original ATLK_irF semantics considering
                   the group as a single agent (distributed knowledge is used)
                 * "individual" for the original ATL_ir semantics considering
                   the group as individual agents (individual knowledge is
                   used)
                 
    If variant is not in {"SF", "FS", "FSF"}, the standard "SF" way is used.          
    If semantics is not in {"group", "individual"}, "group" semantics is used.
    If implem is not in {"naive","generator","optimized","memory","partial",
    "symbolic"}, the standard "generator" way is used.
    
    """
    if implem in __implementations:
        sat = __implementations[implem](mas, spec, variant=variant,
                                        semantics=semantics)
    else:
        sat = evalATLK_gen(mas, spec, variant=variant)
    return (~sat & mas.bddEnc.statesInputsMask & mas.init).is_false()
    
def process(allargs):
    """
    Process specs on the given NuSMV model.
    
    Build the model from a given file and check every ATLK specification
    writen on standard input.
    
    allargs -- a sys.args-like arguments list, without script name.
    """
    
    # Parse arguments
    parser = argparse.ArgumentParser(description='ATLK model checker.')
    # Populate arguments: for now, only the model
    parser.add_argument('model', help='the MAS as an SMV model')
    parser.add_argument('-p', dest='property', help='the property check',
                        default=None)
    parser.add_argument('-v', dest='variant',
                        help='the variant to use (SF, FS, FSF)',
                        default="SF")
    parser.add_argument('-i', dest='implementation',
                        help='the implementation to use (' +
                             ', '.join(__implementations.keys()) + ')',
                        default="naive")
    
    # Debug
    parser.add_argument('-d', dest='debug', help='activate debugging info display',
                        action='store_true', default=False)

    # Config-related arguments
    parser.add_argument('-pe', dest='early',
                        help='activate early termination for partial '
                             'strategies: full, partial or threshold (int) '
                             '(default: None)',
                        default=None)
    parser.add_argument('-pc', dest='caching', help='activate caching',
                        action='store_true', default=False)
    parser.add_argument('-pf', dest='filtering', help='activate filtering',
                        action='store_true', default=False)
    parser.add_argument('-ps', dest='separation',
                        help='activate separation of states for partial '
                             'strategies: random, reach (default: None)',
                        default=None)
    parser.add_argument('-g', dest='garbage',
                        help='activate explicit garbage collection: '
                        'each or step (int) (default: None)', default=None)
    
    
    args = parser.parse_args(allargs)
    
    # Initialize the model
    glob.load_from_file(args.model)
    mas = glob.mas()
    
    # Configure model checking
    config.debug = args.debug
    try:
        config.garbage.step = int(args.garbage)
        config.garbage.type = "step"
    except:
        config.garbage.type = args.garbage
    try:
        config.partial.early.threshold = float(args.early)
        config.partial.early.type = "threshold"
    except:
        config.partial.early.type = args.early
    config.partial.caching = args.caching
    config.partial.filtering = args.filtering
    config.partial.separation.type = args.separation
    
    # Warnings for FS variant with partial implementation
    if args.variant == "FS" and args.implementation == "partial":
        # early termination
        if config.partial.early.type not in {None, "full"}:
            print("Warning: {} early termination is not implemented for "
                  "partial FS variant.".format(str(config.parial.early.type)))
        # filtering
        if not config.partial.filtering:
            print("Warning: filtering is always used for partial FS variant.")
    
    # Check given property, if any
    if args.property:
        try:
            spec = parseATLK(args.property)[0]
            satisfied = check(mas, spec, variant=args.variant,
                              implem=args.implementation)
            print('Specification', str(spec), 'is', str(satisfied))
        except ParseException as e:
            print("[ERROR] Cannot parse specification:", str(e))
        except PyNuSMVError as e:
            print("[ERROR]", str(e))
    
    # Otherwise, ask for ATLK properties
    else:
        print("Enter ATLK properties, one by line:")
        for line in sys.stdin:
            try:
                spec = parseATLK(line)[0]
            
                satisfied = check(mas, spec, variant=args.variant,
                                  implem=args.implementation)
                print('Specification', str(spec), 'is', str(satisfied))
            except ParseException as e:
                print("[ERROR] Cannot parse specification:", str(e))
            except PyNuSMVError as e:
                print("[ERROR]", str(e))

def main():
    with init_nusmv():
        process(sys.argv[1:])

if __name__ == '__main__':
    main()



"""
ARCTL CLI with TLACE explanation.
"""

import cmd
import argparse

from pynusmv.init import init_nusmv, deinit_nusmv, reset_nusmv
from pynusmv.fsm import BddFsm
from pynusmv.exception import PyNuSMVError

from pynusmv_tools.arctl.parsing import parseArctl
from pynusmv_tools.arctl.check import checkArctl
from pynusmv_tools.arctl.tlace.explain import explain_witness, explain_countex
from pynusmv_tools.arctl.tlace.xml import xml_representation

from pyparsing import ParseException


class ARCTL_TLACE_shell(cmd.Cmd):

    prompt = "> "
    
    fsm = None
    fsmpath = None
    
    def preloop(self):
        init_nusmv()
        
    def postloop(self):
        deinit_nusmv()
        
        
    def do_read(self, arg):
        """
        Read file as SMV model to get FSM.
        usage: read FILEPATH
        """
        if len(arg) < 1:
            print("[ERROR] read command needs the SMV model path.")
        else:
            if self.fsm is not None:
                reset_nusmv()
            try:
                self.fsm = BddFsm.from_filename(arg)
                self.fsmpath = arg
            except Exception as e:
                print("[ERROR]", e)
            
            
    def do_check(self, arg):
        """
        Check an ARCTL specification on the current FSM.
        usage: check SPEC
        """
        if self.fsm is None:
            print("[ERROR] No FSM read. Use read command before.")
        elif arg == "":
            print("[ERROR] Need a specification.")
        else:
            try:
                spec = parseArctl(arg)[0]
                (res, wit) = checkArctl(self.fsm, spec,
                                        explain_witness, explain_countex)
                if res:
                    print("The specification", arg, "is true,",
                          "witnessed by")
                    print(xml_representation(self.fsm, wit, spec))
                        
                else:
                    print("The specification", arg, "is false,",
                          "as shown by")
                    print(xml_representation(self.fsm, wit, spec))
                    
            except PyNuSMVError as e:
                print("[ERROR]", e)
                
            except ParseException as e:
                print("[ERROR]", e)
                
                
    def do_fsm(self, arg):
        """
        Print the path to the currently read FSM.
        usage: fsm
        """
        if self.fsmpath is None:
            print("No FSM.")
        else:
            print(self.fsmpath)
        
            
            
    def do_quit(self, arg):
        """
        Quit this prompt.
        usage: quit
        """
        return True
        
        

if __name__ == "__main__":
    ARCTL_TLACE_shell().cmdloop()


"""
ATLK with partial observability evaluation functions.

These evaluation functions reimplement the algorithms and try to be more
memory efficient.
"""

from functools import reduce
import gc

from pynusmv.dd import BDD
from pynusmv.mc import eval_simple_expression
from pynusmv.utils import fixpoint as fp
from pynusmv.exception import PyNuSMVError

from ..atlkFO.ast import (TrueExp, FalseExp, Init, Reachable,
                          Atom, Not, And, Or, Implies, Iff, 
                          AF, AG, AX, AU, AW, EF, EG, EX, EU, EW,
                          nK, nE, nD, nC, K, E, D, C,
                          CEF, CEG, CEX, CEU, CEW, CAF, CAG, CAX, CAU, CAW)

from . import config

from ..atlkFO.eval import (fair_states, ex, eg, eu, nk, ne, nd, nc)


def evalATLK(fsm, spec, variant="SF", semantics="group"):
    """
    Return the BDD representing the set of states of fsm satisfying spec.
    
    fsm -- a MAS representing the system
    spec -- an AST-based ATLK specification
    variant -- the variant of the algorithm to evaluate strategic operators;
               must be
               * "SF" for the standard way: splitting in uniform strategies then
                 filtering winning states,
               * "FS" for the alternating way: filtering winning states, then
                 splitting one conflicting equivalence class, then recurse
               * "FSF" for the filter-split-filter way: filtering winning states
                 then splitting all remaining actions into uniform strategies,
                 then filtering final winning states.
                 
    If variant is not in {"SF", "FS", "FSF"}, the standard "SF" way is used.
    """
    
    if semantics != "group":
        raise PyNuSMVError("Memory evalATLK: unsupported semantics:" +
                           semantics)
    
    if type(spec) is TrueExp:
        return BDD.true(fsm.bddEnc.DDmanager)
        
    elif type(spec) is FalseExp:
        return BDD.false(fsm.bddEnc.DDmanager)
        
    elif type(spec) is Init:
        return fsm.init
        
    elif type(spec) is Reachable:
        return fsm.reachable_states
    
    elif type(spec) is Atom:
        return eval_simple_expression(fsm, spec.value)
        
    elif type(spec) is Not:
        return ~evalATLK(fsm, spec.child, variant=variant)
        
    elif type(spec) is And:
        return (evalATLK(fsm, spec.left, variant=variant)
                & evalATLK(fsm, spec.right, variant=variant))
        
    elif type(spec) is Or:
        return (evalATLK(fsm, spec.left, variant=variant)
                | evalATLK(fsm, spec.right, variant=variant))
        
    elif type(spec) is Implies:
        # a -> b = ~a | b
        return ((~evalATLK(fsm, spec.left, variant=variant))
                 | evalATLK(fsm, spec.right, variant=variant))
        
    elif type(spec) is Iff:
        # a <-> b = (a & b) | (~a & ~b)
        l = evalATLK(fsm, spec.left, variant=variant)
        r = evalATLK(fsm, spec.right, variant=variant)
        return (l & r) | ((~l) & (~r))
        
    elif type(spec) is EX:
        return ex(fsm, evalATLK(fsm, spec.child, variant=variant))
        
    elif type(spec) is AX:
        # AX p = ~EX ~p
        return ~ex(fsm, ~evalATLK(fsm, spec.child, variant=variant))
        
    elif type(spec) is EG:
        return eg(fsm, evalATLK(fsm, spec.child, variant=variant))
        
    elif type(spec) is AG:
        # AG p = ~EF ~p = ~E[ true U ~p ]
        return ~eu(fsm,
                   BDD.true(fsm.bddEnc.DDmanager),
                   ~evalATLK(fsm, spec.child, variant=variant))
        
    elif type(spec) is EU:
        return eu(fsm, evalATLK(fsm, spec.left, variant=variant),
                       evalATLK(fsm, spec.right, variant=variant))
        
    elif type(spec) is AU:
        # A[p U q] = ~E[~q W ~p & ~q] = ~(E[~q U ~p & ~q] | EG ~q)
        p = evalATLK(fsm, spec.left, variant=variant)
        q = evalATLK(fsm, spec.right, variant=variant)
        equpq = eu(fsm, ~q, ~q & ~p)
        egq = eg(fsm, ~q)
        return ~(equpq | egq)
        
    elif type(spec) is EF:
        # EF p = E[ true U p ]
        return eu(fsm,
                  BDD.true(fsm.bddEnc.DDmanager),
                  evalATLK(fsm, spec.child, variant=variant))    
        
    elif type(spec) is AF:
        # AF p = ~EG ~p
        return ~eg(fsm, ~evalATLK(fsm, spec.child, variant=variant))
        
    elif type(spec) is EW:
        # E[ p W q ] = E[ p U q ] | EG p
        return (eu(fsm, evalATLK(fsm, spec.left, variant=variant),
                        evalATLK(fsm, spec.right, variant=variant)) |
                eg(fsm, evalATLK(fsm, spec.left, variant=variant)))
        
    elif type(spec) is AW:
        # A[p W q] = ~E[~q U ~p & ~q]
        p = evalATLK(fsm, spec.left, variant=variant)
        q = evalATLK(fsm, spec.right, variant=variant)
        return ~eu(fsm, ~q, ~p & ~q)
        
    elif type(spec) is nK:
        return nk(fsm, spec.agent.value, evalATLK(fsm, spec.child, variant=variant))
        
    elif type(spec) is K:
        # K<'a'> p = ~nK<'a'> ~p
        return ~nk(fsm, spec.agent.value, ~evalATLK(fsm, spec.child, variant=variant))
        
    elif type(spec) is nE:
        return ne(fsm,
                  [a.value for a in spec.group],
                  evalATLK(fsm, spec.child, variant=variant))
        
    elif type(spec) is E:
        # E<g> p = ~nE<g> ~p
        return ~ne(fsm,
                   [a.value for a in spec.group],
                   ~evalATLK(fsm, spec.child, variant=variant))
        
    elif type(spec) is nD:
        return nd(fsm,
                  [a.value for a in spec.group],
                  evalATLK(fsm, spec.child, variant=variant)) 
        
    elif type(spec) is D:
        # D<g> p = ~nD<g> ~p
        return ~nd(fsm,
                   [a.value for a in spec.group],
                   ~evalATLK(fsm, spec.child, variant=variant))
        
    elif type(spec) is nC:
        return nc(fsm,
                  [a.value for a in spec.group],
                  evalATLK(fsm, spec.child, variant=variant))
        
    elif type(spec) is C:
        # C<g> p = ~nC<g> ~p
        return ~nc(fsm,
                   [a.value for a in spec.group],
                   ~evalATLK(fsm, spec.child, variant=variant))
    
    elif type(spec) is CAX:
        # [g] X p = ~<g> X ~p
        newspec = CEX(spec.group, Not(spec.child))
        return ~evalATLK(fsm, newspec, variant=variant)
        
    elif type(spec) is CAG:
        # [g] G p = ~<g> F ~p
        newspec = CEF(spec.group, Not(spec.child))
        return ~evalATLK(fsm, newspec, variant=variant)
        
    elif type(spec) is CAU:
        # [g][p U q] = ~<g>[ ~q W ~p & ~q ]
        newspec = CEW(spec.group,
                      Not(spec.right),
                      And(Not(spec.left), Not(spec.right)))
        return ~evalATLK(fsm, newspec, variant=variant)
        
    elif type(spec) is CAF:
        # [g] F p = ~<g> G ~p
        newspec = CEG(spec.group, Not(spec.child))
        return ~evalATLK(fsm, newspec, variant=variant)
        
    elif type(spec) is CAW:
        # [g][p W q] = ~<g>[~q U ~p & ~q]
        newspec = CEU(spec.group,
                      Not(spec.right),
                      And(Not(spec.left), Not(spec.right)))
        return ~evalATLK(fsm, newspec, variant=variant)
                   
    elif type(spec) in {CEX, CEG, CEU, CEF, CEW}:
        if variant == "SF":
            return eval_strat(fsm, spec)
        elif variant == "FS":
            print("[WARNING] evalATLK: no improved evaluation for now",
                  " basic evaluation used instead.")
            #return eval_strat_improved(fsm, spec)
            return eval_strat(fsm, spec)
        elif variant == "FSF":
            print("[WARNING] evalATLK: no FSF evaluation for now",
                  " basic evaluation used instead.")
            #return eval_strat_FSF(fsm, spec)
            return eval_strat(fsm, spec)
        else:
            return eval_strat(fsm, spec)
        
    else:
        # TODO Generate error
        print("[ERROR] evalATLK: unrecognized specification type", spec)
        return None


def cex_si(fsm, agents, phi, strat=None):
    """
    Return the set of state/inputs pairs of strat satisfying <agents> X phi
    under full observability in strat.
    If strat is None, strat is considered true.
    
    fsm -- a MAS representing the system
    agents -- a list of agents names
    phi -- a BDD representing the set of states of fsm satisfying phi
    strat -- a BDD representing allowed state/inputs pairs, or None
    """
    if not strat:
        strat = BDD.true(fsm.bddEnc.DDmanager)
    
    phi = phi & fsm.bddEnc.statesInputsMask
    
    return fsm.pre_strat_si(phi | nfair_gamma_si(fsm, agents, strat),
                            agents, strat)
    

def ceu_si(fsm, agents, phi, psi, strat=None):
    """
    Return the set of state/inputs pairs of strat satisfying <agents>[phi U psi]
    under full observability in strat.
    If strat is None, strat is considered true.
    
    fsm -- a MAS representing the system
    agents -- a list of agents names
    phi -- a BDD representing the set of states of fsm satisfying phi
    psi -- a BDD representing the set of states of fsm satisfying psi
    strat -- a BDD representing allowed state/inputs pairs, or None
    
    """
    if not strat:
        strat = BDD.true(fsm.bddEnc.DDmanager)
    
    phi = phi & fsm.bddEnc.statesInputsMask & strat
    psi = psi & fsm.bddEnc.statesInputsMask & strat
    
    if len(fsm.fairness_constraints) == 0:
        return fp(lambda Z : psi | (phi & fsm.pre_strat_si(Z, agents, strat)),
                  BDD.false(fsm.bddEnc.DDmanager))
    else:
        nfair = nfair_gamma_si(fsm, agents, strat)
        def inner(Z):
            res = psi
            for f in fsm.fairness_constraints:
                nf = ~f & fsm.bddEnc.statesMask & strat
                res = res | fsm.pre_strat_si(fp(lambda Y :
                                                 (phi | psi | nfair) &
                                                 (Z | nf) &
                                                 (psi |
                                                  fsm.pre_strat_si(Y, agents,
                                                                   strat)
                                                 ),
                                              BDD.true(fsm.bddEnc.DDmanager)),
                                              agents, strat)
            return (psi | phi | nfair) & res
        return fp(inner, BDD.false(fsm.bddEnc.DDmanager))
    

def cew_si(fsm, agents, phi, psi, strat=None):
    """
    Return the set of state/inputs pairs of strat satisfying <agents>[phi W psi]
    under full observability in strat.
    If strat is None, strat is considered true.
    
    fsm -- a MAS representing the system
    agents -- a list of agents names
    phi -- a BDD representing the set of states of fsm satisfying phi
    psi -- a BDD representing the set of states of fsm satisfying psi
    strat -- a BDD representing allowed state/inputs pairs, or None
    
    """
    if not strat:
        strat = BDD.true(fsm.bddEnc.DDmanager)
    
    phi = phi & fsm.bddEnc.statesInputsMask & strat
    psi = psi & fsm.bddEnc.statesInputsMask & strat
    
    nfair = nfair_gamma_si(fsm, agents, strat)
    
    return fp(lambda Y : (psi | phi | nfair) &
                         (psi | fsm.pre_strat_si(Y, agents, strat)),
              BDD.true(fsm.bddEnc.DDmanager))
    
    
def ceg_si(fsm, agents, phi, strat=None):
    """
    Return the set of state/inputs pairs of strat satisfying <agents> G phi
    under full observability in strat.
    If strat is None, strat is considered true.
    
    fsm -- a MAS representing the system
    agents -- a list of agents names
    phi -- a BDD representing the set of states of fsm satisfying phi
    strat -- a BDD representing allowed state/inputs pairs, or None
    
    """
    if not strat:
        strat = BDD.true(fsm.bddEnc.DDmanager)
    
    phi = phi & fsm.bddEnc.statesInputsMask & strat
    
    nfair = nfair_gamma_si(fsm, agents, strat)
    
    return fp(lambda Y : (phi | nfair) & fsm.pre_strat_si(Y, agents, strat),
              BDD.true(fsm.bddEnc.DDmanager))


def nfair_gamma_si(fsm, agents, strat=None):
    """
    Return the set of state/inputs pairs of strat
    in which agents can avoid a fair path in strat.
    If strat is None, it is considered true.
    
    fsm -- the model
    agents -- a list of agents names
    strat -- a BDD representing allowed state/inputs pairs, or None
    
    """
    if not strat:
        strat = BDD.true(fsm.bddEnc.DDmanager)
    
    if len(fsm.fairness_constraints) == 0:
        return BDD.false(fsm.bddEnc.DDmanager)
    else:
        def inner(Z):
            res = BDD.false(fsm.bddEnc.DDmanager)
            for f in fsm.fairness_constraints:
                nf = ~f & fsm.bddEnc.statesMask & strat
                res = res | fsm.pre_strat_si(fp(lambda Y :
                                                 (Z | nf) &
                                                 fsm.pre_strat_si(Y, agents,
                                                                  strat),
                                             BDD.true(fsm.bddEnc.DDmanager)),
                                             agents, strat)
            return res
        return fp(inner, BDD.false(fsm.bddEnc.DDmanager))


def filter_strat(fsm, spec, strat=None, variant="SF"):
    """
    Returns the subset SA of strat (or the whole system if strat is None),
    state/action pairs of fsm, such that there is a strategy to satisfy spec
    in fsm.
    
    fsm -- a MAS representing the system;
    spec -- an AST-based ATLK specification with a top strategic operator;
            the operator is CEX, CEG, CEF, CEU or CEW.
    strat -- the subset of the system to consider.
    variant -- the variant of the algorithm to evaluate strategic operators;
               must be
               * "SF" for the standard way: splitting in uniform strategies then
                 filtering winning states,
               * "FS" for the alternating way: filtering winning states, then
                 splitting one conflicting equivalence class, then recurse
               * "FSF" for the filter-split-filter way: filtering winning states
                 then splitting all remaining actions into uniform strategies,
                 then filtering final winning states.
                 
    If variant is not in {"SF", "FS", "FSF"}, the standard "SF" way is used.
    """
    
    sat = BDD.false(fsm.bddEnc.DDmanager)
    agents = {atom.value for atom in spec.group}
    
    # Filtering
    if type(spec) is CEX:
        winning = cex_si(fsm, agents,
                         evalATLK(fsm, spec.child, variant=variant), strat)

    elif type(spec) is CEG:
        winning = ceg_si(fsm, agents, evalATLK(fsm, spec.child, variant=variant), strat)

    elif type(spec) is CEU:
        winning = ceu_si(fsm, agents, evalATLK(fsm, spec.left, variant=variant),
                         evalATLK(fsm, spec.right, variant=variant), strat)

    elif type(spec) is CEF:
        # <g> F p = <g>[true U p]
        winning = ceu_si(fsm, agents, BDD.true(fsm.bddEnc.DDmanager),
                         evalATLK(fsm, spec.child, variant=variant), strat)

    elif type(spec) is CEW:
       winning = cew_si(fsm, agents, evalATLK(fsm, spec.left, variant=variant),
                        evalATLK(fsm, spec.right, variant=variant), strat)
    
    
    return winning & fsm.bddEnc.statesInputsMask & fsm.protocol(agents)
    

def all_equiv_sat(fsm, winning, agents):
    """
    Return the states s of winning such that all states indistinguishable
    from s by agents are in winning.
    
    fsm -- a MAS representing the system;
    winning -- a BDD representing a set of states of fsm;
    agents -- a set of agents.
    
    """
    # Get states for which all states belong to winning
    # wineq is the set of states for which all equiv states are in winning
    nwinning = ~winning & fsm.bddEnc.statesInputsMask
    return ~(fsm.equivalent_states(nwinning &
             fsm.reachable_states, frozenset(agents))) & winning


def eval_strat(fsm, spec):
    """
    Return the BDD representing the set of states of fsm satisfying spec.
    spec is a strategic operator <G> pi.
    
    fsm -- a MAS representing the system;
    spec -- an AST-based ATLK specification with a top strategic operator.
    
    """
    
    sat = BDD.false(fsm.bddEnc.DDmanager)
    agents = {atom.value for atom in spec.group}
    gamma_cube = fsm.inputs_cube_for_agents(agents)
    ngamma_cube = fsm.bddEnc.inputsCube - gamma_cube
    
    
    # ----- Produce a first strategy -----
    
    strats = fsm.protocol(agents)
    # Keep track of common part
    common = BDD.false(fsm.bddEnc.DDmanager)
    # Keep track of choice points: done and remaining
    choiced = []
    to_choose = []
    # Build a first strategy (only the current choices)
    current_choices = BDD.false(fsm.bddEnc.DDmanager)
    
    while strats.isnot_false():
        # Get one equivalence class
        si = fsm.pick_one_state_inputs(strats)
        s = si.forsome(fsm.bddEnc.inputsCube)
        eqs = fsm.equivalent_states(s, agents)
        eqcl = strats & eqs
        
        # Remove eqcl from strats
        strats = strats - eqcl
        
        # Get one non-conflicting subset
        ncss = eqcl & si.forsome(fsm.bddEnc.statesCube | ngamma_cube)
        
        # The current equivalence class is conflicting
        if (eqcl - ncss).isnot_false():
            # Add ncss to the current choices and to choiced
            current_choices += ncss
            choiced.append(ncss)
            # Keep the rest in to_choose
            to_choose.append(eqcl - ncss)
        else:
            # Add equivalence class to common
            common += eqcl
    
    # ---- Model check the current strategy (current_choices + common) -----
    
    winning = (filter_strat(fsm, spec, current_choices + common, "SF")
               .forsome(fsm.bddEnc.inputsCube))
    sat = all_equiv_sat(fsm, winning, agents)
    
    # If no choices have been made, the full protocol is non-conflicting
    # and we already checked to only one strategy
    if len(to_choose) <= 0:
        return sat
    
    # At this point:
    # - common contains the part common to all strategies:
    #   equivalence classes where no choice had to be made
    # - current_choices contains the rest of the current strategy:
    #   equivalence classes where choices have been made
    #   => non-conflicting subsets
    # - choiced contains the list of choices made
    # - to_choose contains the list of remaining choices to make
    # - sat contains the states for which the first strategy
    #   (current_choices + common) is winning
    
    
    # ----- Test all remaining strategies -----
    nbstrats = 1
    while (reduce(lambda x,y: x | y, to_choose,
                  BDD.false(fsm.bddEnc.DDmanager))).isnot_false():
        # If some to_choose element is not false, we still have choices to try
        
        # --- Build the next strategy ---
        
        choice = len(to_choose) - 1
        while to_choose[choice].is_false():
            # Stop when we reach a choice where we still have to choose
            # Guaranteed to happen since to_choose elements are not all false
            
            # Shift the completely made choices
            to_choose[choice] = choiced[choice]
            
            # Choose one action
            # Get one equivalence class and one non-conflicting subset
            si = fsm.pick_one_state_inputs(to_choose[choice])
            ncss = (to_choose[choice]
                    & si.forsome(fsm.bddEnc.statesCube | ngamma_cube))
            
            # Update the strategy
            current_choices -= to_choose[choice]
            current_choices += ncss
            
            # Update choices
            choiced[choice] = ncss
            to_choose[choice] -= ncss
            
            choice -= 1
            
        # Shift the current choice and update the strategy
        # to_choose[choice] is not false
        
        # Choose one action
        # Get one equivalence class and one non-conflicting subset
        si = fsm.pick_one_state_inputs(to_choose[choice])
        ncss = (to_choose[choice]
                & si.forsome(fsm.bddEnc.statesCube | ngamma_cube))
        
        # Update the strategy
        # to_choose[choice] + choiced[choice] contains the full equiv class
        current_choices -= to_choose[choice] + choiced[choice]
        current_choices += ncss
        
        # Update choices
        choiced[choice] += ncss
        to_choose[choice] -= ncss
        
        
        # --- Model check the strategy ---
        
        winning = (filter_strat(fsm, spec, current_choices + common, "SF")
                   .forsome(fsm.bddEnc.inputsCube))
        sat += all_equiv_sat(fsm, winning, agents)
        
        gc.collect()
        
        nbstrats += 1
    
    if config.debug:
        print("[DEBUG] Eval strat Mem: {} strats checked".format(nbstrats))
    return sat


import unittest

from pynusmv.init import init_nusmv, deinit_nusmv
from pynusmv.mc import eval_simple_expression
from pynusmv.dd import BDD
from pynusmv import model as md

from pynusmv_tools.mas import glob
from pynusmv_tools.mas.mas import Agent, Group

class TestMAS(unittest.TestCase):
    
    def setUp(self):
        init_nusmv()
        
    def tearDown(self):
        glob.reset_globals()
        deinit_nusmv()
        
    def model(self):
        glob.load_from_file("tests/pynusmv_tools/mas/dining-crypto.smv")
        fsm = glob.mas()
        self.assertIsNotNone(fsm)
        return fsm
        
    
    def cardgame(self):
        glob.load_from_file("tests/pynusmv_tools/mas/cardgame.smv")
        fsm = glob.mas()
        self.assertIsNotNone(fsm)
        return fsm
        
        
    def simple(self):
        glob.load_from_file("tests/pynusmv_tools/mas/simple.smv")
        fsm = glob.mas()
        self.assertIsNotNone(fsm)
        return fsm
        
        
    def premod(self):
        glob.load_from_file("tests/pynusmv_tools/mas/pre.smv")
        fsm = glob.mas()
        self.assertIsNotNone(fsm)
        return fsm
        
        
    def test_pre(self):
        fsm = self.model()
        
        c1p = eval_simple_expression(fsm, "c1.payer")
        unknown = eval_simple_expression(fsm, "countsay = unknown")
        odd = eval_simple_expression(fsm, "countsay = odd")
        even = eval_simple_expression(fsm, "countsay = even")
        true = eval_simple_expression(fsm, "TRUE")
        
        self.assertTrue(odd & fsm.bddEnc.statesMask <= fsm.pre(odd))
        self.assertTrue(fsm.init <= unknown)
        self.assertTrue(fsm.pre(unknown).is_false())
        
        
    def test_simple_pre(self):
        glob.load_from_file("tests/pynusmv_tools/ctlk/agents.smv")
        fsm = glob.mas()
        self.assertIsNotNone(fsm)
        
        lt = eval_simple_expression(fsm, "at.local")
        lf = eval_simple_expression(fsm, "af.local")
        g = eval_simple_expression(fsm, "global")
        true = eval_simple_expression(fsm, "TRUE")
        false = eval_simple_expression(fsm, "FALSE")
        
        self.assertEqual(fsm.pre(lt & ~lf & ~g),
                         (lt & lf & ~g) | (~lt & ~lf & g))
        self.assertEqual(fsm.pre(g), true)
        self.assertEqual(fsm.pre(lf), lf.iff(g))
        
        
    def test_count(self):
        fsm = self.model()
        
        c1p = eval_simple_expression(fsm, "c1.payer")
        c2p = eval_simple_expression(fsm, "c2.payer")
        c3p = eval_simple_expression(fsm, "c3.payer")
        c1ch = eval_simple_expression(fsm, "c1.coin = head")
        c2ch = eval_simple_expression(fsm, "c2.coin = head")
        c3ch = eval_simple_expression(fsm, "c3.coin = head")
        c1se = eval_simple_expression(fsm, "c1.say = equal")
        c2se = eval_simple_expression(fsm, "c2.say = equal")
        c3se = eval_simple_expression(fsm, "c3.say = equal")        
        unknown = eval_simple_expression(fsm, "countsay = unknown")
        odd = eval_simple_expression(fsm, "countsay = odd")
        even = eval_simple_expression(fsm, "countsay = even")
        true = eval_simple_expression(fsm, "TRUE")
        false = eval_simple_expression(fsm, "FALSE")
        
        self.assertEqual(1, fsm.count_states(fsm.pick_one_state(c1p)))
        
        self.assertEqual(1, fsm.count_states(c1p & ~c2p & ~c3p & c1ch & c2ch & c3ch & unknown))
        self.assertEqual(1, fsm.count_states(c1p & ~c2p & ~c3p & c1ch & c2ch & c3ch & odd))
        
        
    def test_constraints_post(self):
        glob.load_from_file("tests/pynusmv_tools/ctlk/constraints.smv")
        fsm = glob.mas()
        self.assertIsNotNone(fsm)
        
        false = eval_simple_expression(fsm, "FALSE")
        false = eval_simple_expression(fsm, "TRUE")
        p = eval_simple_expression(fsm, "p")
        q = eval_simple_expression(fsm, "q")
        a = eval_simple_expression(fsm, "a")
        
        self.assertEqual(1, fsm.count_states(fsm.init))
        self.assertEqual(2, fsm.count_states(fsm.post(fsm.init)))
        self.assertEqual(1, fsm.count_states(p & q))
        self.assertEqual(1, fsm.count_states(p & ~q))
        self.assertEqual(1, fsm.count_states(~p & q))
        self.assertEqual(1, fsm.count_states(~p & ~q))
        
        self.assertEqual(2, fsm.count_states(fsm.post(p & q)))
        self.assertEqual(1, fsm.count_states(fsm.post(p & ~q)))
        self.assertEqual(1, fsm.count_states(fsm.post(~p & q)))
        self.assertEqual(1, fsm.count_states(fsm.post(p & q, a)))
        self.assertEqual(0, fsm.count_states(fsm.post(p & ~q, ~a)))
        self.assertEqual(0, fsm.count_states(fsm.post(~p & q, a)))
        self.assertEqual(1, fsm.count_states(fsm.post(~p & q, ~a)))
                
        
    def test_simple_post(self):
        glob.load_from_file("tests/pynusmv_tools/ctlk/agents.smv")
        fsm = glob.mas()
        self.assertIsNotNone(fsm)
        
        lt = eval_simple_expression(fsm, "at.local")
        lf = eval_simple_expression(fsm, "af.local")
        g = eval_simple_expression(fsm, "global")
        true = eval_simple_expression(fsm, "TRUE")
        false = eval_simple_expression(fsm, "FALSE")
        
        self.assertEqual(fsm.post(lt & lf & g), ~lt & lf & g | ~lt & lf & ~g)
        self.assertEqual(fsm.post(lt & g), ~lt)
        self.assertEqual(fsm.post(lt & ~g), lt)
        self.assertEqual(fsm.post(lt & lf), lt.iff(~lf))
        
        
    def test_equivalent_states(self):
        fsm = self.model()
        
        c1p = eval_simple_expression(fsm, "c1.payer")
        odd = eval_simple_expression(fsm, "countsay = odd")
        true = eval_simple_expression(fsm, "TRUE")
        
        self.assertEqual(fsm.equivalent_states(c1p, {"c1"}), c1p)
        self.assertEqual(fsm.equivalent_states(c1p, {"c2"}), true)
        
        
    def test_reachable_states_for_simple_model(self):
        glob.load_from_file("tests/pynusmv_tools/ctlk/agents.smv")
        fsm = glob.mas()
        self.assertIsNotNone(fsm)
        
        true = eval_simple_expression(fsm, "TRUE")
        self.assertEqual(fsm.reachable_states, true)
        
        
    def test_reachable_states(self):
        fsm = self.model()
        
        c1p = eval_simple_expression(fsm, "c1.payer")
        c2p = eval_simple_expression(fsm, "c2.payer")
        odd = eval_simple_expression(fsm, "countsay = odd")
        true = eval_simple_expression(fsm, "TRUE")
        
        self.assertTrue(fsm.reachable_states.isnot_true())
        self.assertTrue(fsm.reachable_states.isnot_false())
        self.assertTrue((fsm.init & c1p & c2p).is_false())
        self.assertTrue((fsm.post(fsm.init) & c1p & c2p).is_false())
        
        tmp = fsm.reachable_states & (c1p & c2p)
        while tmp.isnot_false():
            s = fsm.pick_one_state(tmp)
            print(s.get_str_values())
            tmp -= s
        
        self.assertTrue((fsm.reachable_states & (c1p & c2p)).is_false())
        
        
    def test_agents(self):
        fsm = self.model()
        self.assertSetEqual({'c1','c2','c3'}, fsm.agents)
        
        
    def test_observed_variables(self):
        fsm = self.model()
        
        self.assertSetEqual({'c1','c2','c3'},
                            set(fsm.agents_observed_variables.keys()))
        
        self.assertSetEqual(fsm.agents_observed_variables['c1'],
                            {'c1.coin', 'c1.payer', 'c3.coin', 'countsay'})
        self.assertSetEqual(fsm.agents_observed_variables['c2'],
                            {'c2.coin', 'c2.payer', 'c1.coin', 'countsay'})
        self.assertSetEqual(fsm.agents_observed_variables['c3'],
                            {'c3.coin', 'c3.payer', 'c2.coin', 'countsay'})
        
        
    def test_input_variables(self):
        fsm = self.model()
        
        self.assertSetEqual({'c1','c2','c3'}, set(fsm.agents_inputvars.keys()))
        
        self.assertSetEqual(fsm.agents_inputvars['c1'], {'c1.say'})
        self.assertSetEqual(fsm.agents_inputvars['c2'], {'c2.say'})
        self.assertSetEqual(fsm.agents_inputvars['c3'], {'c3.say'})
        
        
    def test_protocol(self):
        fsm = self.cardgame()
        
        s0 = eval_simple_expression(fsm, "step = 0")
        s1 = eval_simple_expression(fsm, "step = 1")
        s2 = eval_simple_expression(fsm, "step = 2")
        pa = eval_simple_expression(fsm, "pcard = Ac")
        pk = eval_simple_expression(fsm, "pcard = K")
        pq = eval_simple_expression(fsm, "pcard = Q")
        da = eval_simple_expression(fsm, "dcard = Ac")
        dk = eval_simple_expression(fsm, "dcard = K")
        dq = eval_simple_expression(fsm, "dcard = Q")
        dda = eval_simple_expression(fsm, "ddcard = Ac")
        ddk = eval_simple_expression(fsm, "ddcard = K")
        ddq = eval_simple_expression(fsm, "ddcard = Q")
        win = eval_simple_expression(fsm, "win")
        lose = eval_simple_expression(fsm, "lose")
        true = eval_simple_expression(fsm, "TRUE")
        false = eval_simple_expression(fsm, "FALSE")
        dak = eval_simple_expression(fsm, "dealer.action = dealAK")
        dn = eval_simple_expression(fsm, "dealer.action = none")
        ps = eval_simple_expression(fsm, "player.action = swap")
        pk = eval_simple_expression(fsm, "player.action = keep")
        pn = eval_simple_expression(fsm, "player.action = none")
        
        self.assertTrue((s0 & dak & fsm.protocol({"dealer"})).isnot_false())
        self.assertTrue((s0 & dak & fsm.reachable_states &
                         fsm.bddEnc.statesInputsMask) <= fsm.protocol({"dealer"}))
        self.assertFalse((s0 & dn & fsm.reachable_states &
                         fsm.bddEnc.statesInputsMask) <= fsm.protocol({"dealer"}))
        
        self.assertTrue((s0 & pn & fsm.reachable_states &
                         fsm.bddEnc.statesInputsMask) <= fsm.protocol({"player"}))
        self.assertTrue((s1 & pk & fsm.reachable_states) &
                         fsm.bddEnc.statesInputsMask <= fsm.protocol({"player"}))
        self.assertTrue((s2 & pn & fsm.reachable_states &
                         fsm.bddEnc.statesInputsMask) <= fsm.protocol({"player"}))
        self.assertTrue((s1 & ps & fsm.reachable_states &
                         fsm.bddEnc.statesInputsMask) <= fsm.protocol({"player"}))
        self.assertFalse((s1 & pn & fsm.reachable_states &
                         fsm.bddEnc.statesInputsMask) <= fsm.protocol({"player"}))
    
    
    def test_pre_strat(self):
        fsm = self.cardgame()
        
        s0 = eval_simple_expression(fsm, "step = 0")
        s1 = eval_simple_expression(fsm, "step = 1")
        s2 = eval_simple_expression(fsm, "step = 2")
        pa = eval_simple_expression(fsm, "pcard = Ac")
        pk = eval_simple_expression(fsm, "pcard = K")
        pq = eval_simple_expression(fsm, "pcard = Q")
        da = eval_simple_expression(fsm, "dcard = Ac")
        dk = eval_simple_expression(fsm, "dcard = K")
        dq = eval_simple_expression(fsm, "dcard = Q")
        dda = eval_simple_expression(fsm, "ddcard = Ac")
        ddk = eval_simple_expression(fsm, "ddcard = K")
        ddq = eval_simple_expression(fsm, "ddcard = Q")
        win = eval_simple_expression(fsm, "win")
        lose = eval_simple_expression(fsm, "lose")
        true = eval_simple_expression(fsm, "TRUE")
        false = eval_simple_expression(fsm, "FALSE")
        
        self.assertTrue(pa & dq & s2 & fsm.reachable_states  & fsm.bddEnc.statesInputsMask <= fsm.pre_strat(pa & dq & s2 & fsm.reachable_states, {'player'}))
        self.assertTrue(pa & dq & s1 & fsm.reachable_states  & fsm.bddEnc.statesInputsMask<= fsm.pre_strat(pa & dq & s2 & fsm.reachable_states, {'player'}))
        self.assertTrue(pk & dq & s1 & fsm.reachable_states  & fsm.bddEnc.statesInputsMask<= fsm.pre_strat(pa & dq & s2 & fsm.reachable_states, {'player'}))
        self.assertFalse(pk & da & s1 & fsm.reachable_states  & fsm.bddEnc.statesInputsMask<= fsm.pre_strat(pa & dq & s2 & fsm.reachable_states, {'player'}))
        
        self.assertTrue(s1 & fsm.reachable_states  & fsm.bddEnc.statesInputsMask<=
                        fsm.pre_strat(win & fsm.reachable_states, {'player'}))
        self.assertTrue(s1 & fsm.reachable_states  & fsm.bddEnc.statesInputsMask<=
                                                fsm.pre_strat(lose, {'player'}))
        
        self.assertTrue(fsm.init & fsm.bddEnc.statesInputsMask <= fsm.pre_strat(s1, {'dealer'}))
        self.assertTrue(fsm.init & fsm.bddEnc.statesInputsMask <= fsm.pre_strat(s1 & pq & da, {'dealer'}))
        
        self.assertFalse(fsm.init & fsm.bddEnc.statesInputsMask <= fsm.pre_strat(s1 & pq & da, {'player'}))
        
    
    def test_simple(self):
        mas = self.simple()
        s4 = eval_simple_expression(mas, "s = 4")
        ac = mas.bddEnc.cube_for_inputs_vars("a")
        bc = mas.bddEnc.cube_for_inputs_vars("b")
        self.assertTrue((~(mas.weak_pre(~s4).forsome(bc)) &     
                        mas.weak_pre(s4)).forsome(mas.bddEnc.inputsCube))
                        
                        
    def test_pre_nstat(self):
        fsm = self.cardgame()
        
        s0 = eval_simple_expression(fsm, "step = 0")
        s1 = eval_simple_expression(fsm, "step = 1")
        s2 = eval_simple_expression(fsm, "step = 2")
        pa = eval_simple_expression(fsm, "pcard = Ac")
        pk = eval_simple_expression(fsm, "pcard = K")
        pq = eval_simple_expression(fsm, "pcard = Q")
        da = eval_simple_expression(fsm, "dcard = Ac")
        dk = eval_simple_expression(fsm, "dcard = K")
        dq = eval_simple_expression(fsm, "dcard = Q")
        dda = eval_simple_expression(fsm, "ddcard = Ac")
        ddk = eval_simple_expression(fsm, "ddcard = K")
        ddq = eval_simple_expression(fsm, "ddcard = Q")
        win = eval_simple_expression(fsm, "win")
        lose = eval_simple_expression(fsm, "lose")
        true = eval_simple_expression(fsm, "TRUE")
        false = eval_simple_expression(fsm, "FALSE")
        
        self.assertTrue(s0 & fsm.reachable_states <=
                        fsm.pre_nstrat(s1 & pa, {"player"}))
        self.assertTrue(s1 & fsm.reachable_states <=
                        fsm.pre_nstrat(win, {"dealer"}))
        self.assertFalse(s1 & fsm.reachable_states <=
                        fsm.pre_nstrat(win, {"player"}))
                        
                        
    def test_pre_stat_si(self):
        fsm = self.cardgame()
        
        s0 = eval_simple_expression(fsm, "step = 0")
        s1 = eval_simple_expression(fsm, "step = 1")
        s2 = eval_simple_expression(fsm, "step = 2")
        pa = eval_simple_expression(fsm, "pcard = Ac")
        pk = eval_simple_expression(fsm, "pcard = K")
        pq = eval_simple_expression(fsm, "pcard = Q")
        da = eval_simple_expression(fsm, "dcard = Ac")
        dk = eval_simple_expression(fsm, "dcard = K")
        dq = eval_simple_expression(fsm, "dcard = Q")
        dda = eval_simple_expression(fsm, "ddcard = Ac")
        ddk = eval_simple_expression(fsm, "ddcard = K")
        ddq = eval_simple_expression(fsm, "ddcard = Q")
        pan = eval_simple_expression(fsm, "player.action = none")
        pak = eval_simple_expression(fsm, "player.action = keep")
        pas = eval_simple_expression(fsm, "player.action = swap")
        daqa = eval_simple_expression(fsm, "dealer.action = dealQA")
        win = eval_simple_expression(fsm, "win")
        lose = eval_simple_expression(fsm, "lose")
        true = eval_simple_expression(fsm, "TRUE")
        false = eval_simple_expression(fsm, "FALSE")
        
        self.assertTrue(pa & dq & s2 & pan & fsm.reachable_states & fsm.bddEnc.statesInputsMask <= fsm.pre_strat_si(pa & dq & s2 & fsm.reachable_states, {'player'}))
        self.assertTrue(pa & dq & s1 & pak & fsm.reachable_states & fsm.bddEnc.statesInputsMask <= fsm.pre_strat_si(pa & dq & s2 & fsm.reachable_states, {'player'}))
        self.assertTrue(pk & dq & s1 & pas & fsm.reachable_states & fsm.bddEnc.statesInputsMask <= fsm.pre_strat_si(pa & dq & s2 & fsm.reachable_states, {'player'}))
        self.assertFalse(pk & dq & s1 & pak & fsm.reachable_states & fsm.bddEnc.statesInputsMask <= fsm.pre_strat_si(pa & dq & s2 & fsm.reachable_states, {'player'}))
        self.assertFalse(pk & da & s1 & fsm.reachable_states & fsm.bddEnc.statesInputsMask <= fsm.pre_strat_si(pa & dq & s2 & fsm.reachable_states, {'player'}))
        
        self.assertFalse(s1 & fsm.reachable_states & fsm.bddEnc.statesInputsMask <=
                         fsm.pre_strat_si(win & fsm.reachable_states, {'player'}))
        self.assertFalse(s1 & fsm.reachable_states & fsm.bddEnc.statesInputsMask <=
                         fsm.pre_strat_si(lose, {'player'}))
                         
        self.assertTrue(win & fsm.reachable_states & fsm.bddEnc.statesInputsMask & fsm.protocol({'player'}) <=
                        fsm.pre_strat_si(win & fsm.reachable_states, {'player'}))
        self.assertTrue(s1 & pk & dq & pas & fsm.reachable_states & fsm.bddEnc.statesInputsMask <=
                        fsm.pre_strat_si(lose, {'player'}))
        
        self.assertTrue(fsm.init & fsm.protocol({'dealer'}) & fsm.bddEnc.statesInputsMask <= fsm.pre_strat_si(s1, {'dealer'}))
        self.assertTrue(fsm.init & daqa & fsm.bddEnc.statesInputsMask <= fsm.pre_strat_si(s1 & pq & da, {'dealer'}))
        
        
    def test_premod_pre_strat_si(self):
        fsm = self.premod()
        
        p = eval_simple_expression(fsm, "a.p = 1")
        q = eval_simple_expression(fsm, "b.q = 1")
        pa = eval_simple_expression(fsm, "a.a = 1")
        qa = eval_simple_expression(fsm, "b.a = 1")
        
        aa = {'a'}
        bb = {'b'}
        ag = {'a', 'b'}
        
        self.assertEqual((~p & q) | (p & ~q & ~pa), fsm.pre_strat_si(~p & q, aa))
        self.assertTrue(fsm.pre_strat_si(p, ag).is_false())
        self.assertEqual(fsm.pre_strat_si(~p & q, bb), (~p & q) | (p & q & ~qa))
        
        strat = (pa)
        
        self.assertEqual((~p & q & pa), fsm.pre_strat_si(~p & q, aa, strat))
        self.assertEqual(fsm.pre_strat_si(~p & q, bb, strat),
                         ((~p & q) | (p & q & ~qa)) & strat)
        self.assertEqual(fsm.pre_strat_si(~p & q, bb, qa),
                         (~p & q & qa))
    
    def test_post(self):
        fsm = self.cardgame()
        
        s0 = eval_simple_expression(fsm, "step = 0")
        s1 = eval_simple_expression(fsm, "step = 1")
        s2 = eval_simple_expression(fsm, "step = 2")
        pa = eval_simple_expression(fsm, "pcard = Ac")
        pk = eval_simple_expression(fsm, "pcard = K")
        pq = eval_simple_expression(fsm, "pcard = Q")
        da = eval_simple_expression(fsm, "dcard = Ac")
        dk = eval_simple_expression(fsm, "dcard = K")
        dq = eval_simple_expression(fsm, "dcard = Q")
        dda = eval_simple_expression(fsm, "ddcard = Ac")
        ddk = eval_simple_expression(fsm, "ddcard = K")
        ddq = eval_simple_expression(fsm, "ddcard = Q")
        win = eval_simple_expression(fsm, "win")
        lose = eval_simple_expression(fsm, "lose")
        true = eval_simple_expression(fsm, "TRUE")
        false = eval_simple_expression(fsm, "FALSE")
        pan = eval_simple_expression(fsm, "player.action = none")
        pak = eval_simple_expression(fsm, "player.action = keep")
        pas = eval_simple_expression(fsm, "player.action = swap")
        daqa = eval_simple_expression(fsm, "dealer.action = dealQA")
        
        self.assertEqual(fsm.post(s0 & fsm.reachable_states) &
                         fsm.bddEnc.statesMask,
                         s1 & fsm.reachable_states & fsm.bddEnc.statesMask)
        self.assertEqual(fsm.post(s0 & fsm.reachable_states, inputs=daqa) &
                         fsm.bddEnc.statesMask,
                         s1 & pq & da & fsm.reachable_states &
                         fsm.bddEnc.statesMask)
        self.assertEqual(fsm.post(s0 & fsm.reachable_states,
                                  subsystem=s0 & daqa) & fsm.bddEnc.statesMask,
                         s1 & pq & da & fsm.reachable_states &
                         fsm.bddEnc.statesMask)
        self.assertEqual(fsm.post(fsm.reachable_states,
                                  subsystem=(s0 & daqa) | (s1 & pak)) &
                         fsm.bddEnc.statesMask,
                         ((s1 & pq & da) | (s2)) & fsm.reachable_states &
                         fsm.bddEnc.statesMask)
        self.assertEqual(fsm.post(s1 & fsm.reachable_states,
                        subsystem=(s1 & pa & dq & pak) | (s1 & pk & da & pas))
                         & fsm.bddEnc.statesMask,
                         ((s2 & pa & dq) | (s2 & da & pq)) &
                         fsm.reachable_states &
                         fsm.bddEnc.statesMask)
    
    def test_cardgame_with_agents(self):
        glob.load_from_file("tests/pynusmv_tools/mas/cardgame.smv")
        agents = (Agent("player", {"step", "pcard", "ddcard"},
                        {"player.action"}),
                  Agent("dealer", {"step", "pcard", "dcard"},
                        {"dealer.action"}))
        fsm = glob.mas(agents=agents)
        self.assertIsNotNone(fsm)
        return fsm
    
    def test_check_mas_free_choice_ok(self):
        class agent(md.Module):
            v = md.Var(md.Boolean())
            a = md.IVar(md.Boolean())
            INIT = v
            TRANS = (v.next() == a)
        class main(md.Module):
            a1 = md.Var(agent())
            a2 = md.Var(agent())
        
        agents = {Agent("a1", {main.a1.v, main.a2.v}, {main.a1.a}),
                  Agent("a2", {main.a1.v, main.a2.v}, {main.a2.a})}
        
        glob.load(agent, main)
        mas = glob.mas(agents=agents)
        self.assertIsNotNone(mas)
        self.assertTrue(mas.check_free_choice().is_false())
    
    def test_check_mas_free_choice_ko(self):
        class agent(md.Module):
            v = md.Var(md.Boolean())
            a = md.IVar(md.Boolean())
            INIT = v
            TRANS = (v.next() == a)
        class main(md.Module):
            a1 = md.Var(agent())
            a2 = md.Var(agent())
            TRANS = a2.a.implies(a1.a)
        
        agents = {Agent("a1", {main.a1.v, main.a2.v}, {main.a1.a}),
                  Agent("a2", {main.a1.v, main.a2.v}, {main.a2.a})}
        
        glob.load(agent, main)
        mas = glob.mas(agents=agents)
        self.assertIsNotNone(mas)
        
        free_choice = mas.check_free_choice()
        self.assertTrue(free_choice.isnot_false())
    
    def test_check_mas_uniform_choice_ok(self):
        class agent(md.Module):
            v = md.Var(md.Boolean())
            a = md.IVar(md.Boolean())
            INIT = v
            TRANS = (v.next() == a)
        class main(md.Module):
            a1 = md.Var(agent())
            a2 = md.Var(agent())
        
        agents = {Agent("a1", {main.a1.v}, {main.a1.a}),
                  Agent("a2", {main.a2.v}, {main.a2.a})}
        
        glob.load(agent, main)
        mas = glob.mas(agents=agents)
        self.assertIsNotNone(mas)
        self.assertTrue(mas.check_uniform_choice().is_false())
    
    def test_check_mas_uniform_choice_ko(self):
        class agent(md.Module):
            v = md.Var(md.Boolean())
            a = md.IVar(md.Boolean())
            INIT = v
            TRANS = (v.next() == a)
        class main(md.Module):
            a1 = md.Var(agent())
            a2 = md.Var(agent())
            TRANS = (a1.a == a2.v)
        
        agents = {Agent("a1", {main.a1.v}, {main.a1.a}),
                  Agent("a2", {main.a2.v}, {main.a2.a})}
        
        glob.load(agent, main)
        mas = glob.mas(agents=agents)
        self.assertIsNotNone(mas)
        
        uniform_choice = mas.check_uniform_choice()
        self.assertTrue(uniform_choice.isnot_false())